from ctypes import POINTER, byref, c_char_p, c_uint8, c_uint16, c_uint32, c_void_p
from enum import IntFlag
from typing import NewType

from koda import Err, Ok, Result

from .. import Ft4222Exception, Ft4222Status, FtHandle
from ..dll_loader import ftlib

I2cMasterHandle = NewType("I2cMasterHandle", FtHandle)


class TransactionFlag(IntFlag):
    """Enum representing possible special transaction flags.

    This enum is used in functions 'read_ex()' and 'write_ex()'.
    """

    NONE = 0x80
    """No special transaction condition."""
    START = 0x02
    """A start condition is signaled."""
    REPEATED_START = 0x03  # Repeated_START will not send master code in HS mode
    """A repeated start condition is signaled."""
    STOP = 0x04
    """A stop condition is signaled."""
    START_AND_STOP = 0x06  # START condition followed by SEND and STOP condition
    """A combination of start and stop conditions is signaled."""


class CtrlStatus(IntFlag):
    """Enum representing controller status flags."""

    CONTROLLER_BUSY = 1 << 0
    """Controller busy. All other status bits are invalid."""
    ERROR = 1 << 1
    """Error condition."""
    SLAVE_ADDR_NACK = 1 << 2
    """Slave address was not acknowledged during the last operation."""
    DATA_NACK = 1 << 3
    """Data not acknowledged during the last operation."""
    ARBITRATION_LOST = 1 << 4
    """Arbitration lost during the last operation."""
    IDLE = 1 << 5
    """Controller idle."""
    BUS_BUSY = 1 << 6
    """Bus busy."""


_init = ftlib.FT4222_I2CMaster_Init
_init.argtypes = [c_void_p, c_uint32]
_init.restype = Ft4222Status

_read = ftlib.FT4222_I2CMaster_Read
_read.argtypes = [c_void_p, c_uint16, POINTER(c_uint8), c_uint16, POINTER(c_uint16)]
_read.restype = Ft4222Status

_write = ftlib.FT4222_I2CMaster_Write
_write.argtypes = [c_void_p, c_uint16, c_char_p, c_uint16, POINTER(c_uint16)]
_write.restype = Ft4222Status

_read_ex = ftlib.FT4222_I2CMaster_ReadEx
_read_ex.argtypes = [
    c_void_p,
    c_uint16,
    c_uint8,
    POINTER(c_uint8),
    c_uint16,
    POINTER(c_uint16),
]
_read_ex.restype = Ft4222Status

_write_ex = ftlib.FT4222_I2CMaster_WriteEx
_write_ex.argtypes = [
    c_void_p,
    c_uint16,
    c_uint8,
    c_char_p,
    c_uint16,
    POINTER(c_uint16),
]
_write_ex.restype = Ft4222Status

_reset = ftlib.FT4222_I2CMaster_Reset
_reset.argtypes = [c_void_p]
_reset.restype = Ft4222Status

_get_status = ftlib.FT4222_I2CMaster_GetStatus
_get_status.argtypes = [c_void_p, POINTER(c_uint8)]
_get_status.restype = Ft4222Status

_reset_bus = ftlib.FT4222_I2CMaster_ResetBus
_reset_bus.argtypes = [c_void_p]
_reset_bus.restype = Ft4222Status


def init(ft_handle: FtHandle, kbps: int) -> Result[I2cMasterHandle, Ft4222Status]:
    """Initialize the FT4222H as an I2C master with the requested I2C speed.

    Args:
        ft_handle:          Handle to an opened FT4222 device
        kbps:               I2C transmission speed (60K - 3400K)

    Returns:
        Result:    Handle to initialized FT4222 device in I2C Master mode
    """
    assert 60 <= kbps <= 3400, "I2C transmission speed must be in range 60 to 3400"

    result: Ft4222Status = _init(ft_handle, kbps)

    if result == Ft4222Status.OK:
        return Ok(I2cMasterHandle(ft_handle))
    else:
        return Err(result)


def read(ft_handle: I2cMasterHandle, dev_address: int, read_byte_count: int) -> bytes:
    """Read data from the specified I2C slave device with START and STOP conditions.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in I2C Master mode
        dev_address:        Address of the target I2C slave
        read_byte_count:    Positive number of bytes to read

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        bytes:              Read data
    """
    assert (
        0 <= dev_address < (2 ** 16)
    ), "Device address must be an 16b unsigned integer (range 0 - 65 535)"
    assert (
        0 < read_byte_count < (2 ** 16)
    ), "Number of bytes to read must be positive and less than 2^16"

    read_buffer = (c_uint8 * read_byte_count)()
    bytes_read = c_uint16()

    result: Ft4222Status = _read(
        ft_handle, dev_address, read_buffer, len(read_buffer), byref(bytes_read)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes(read_buffer[: bytes_read.value])


def write(ft_handle: I2cMasterHandle, dev_address: int, write_data: bytes) -> int:
    """Write data to the specified I2C slave device with START and STOP conditions.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Master mode
        dev_address:    Address of the target I2C slave
        write_data:     List of non-zero length containing bytes to write

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:            Number of bytes written
    """
    assert (
        0 <= dev_address < (2 ** 16)
    ), "Device address must be an 16b unsigned integer (range 0 - 65 535)"
    assert (
        0 < len(write_data) < (2 ** 16)
    ), "Data to be written must be non-empty and contain less than 2^16 bytes"

    bytes_written = c_uint16()

    result: Ft4222Status = _write(
        ft_handle, dev_address, write_data, len(write_data), byref(bytes_written)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes_written.value


def read_ex(
    ft_handle: I2cMasterHandle,
    dev_address: int,
    flag: TransactionFlag,
    read_byte_count: int,
) -> bytes:
    """Read data from the specified I2C slave device with the specified I2C condition.

    NOTE: This function is supported by the rev. B FT4222H or later!

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Master mode
        dev_address:        Address of target I2C slave device
        flag:               I2C transaction condition flag
        read_byte_count:    Positive number of bytes to read

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        bytes:              Read data
    """
    assert (
        0 <= dev_address < (2 ** 16)
    ), "Device address must be an 16b unsigned integer (range 0 - 65 535)"
    assert (
        0 < read_byte_count < (2 ** 16)
    ), "Number of bytes to read must be positive and less than 2^16"

    read_buffer = (c_uint8 * read_byte_count)()
    bytes_read = c_uint16()

    result: Ft4222Status = _read_ex(
        ft_handle, dev_address, flag, read_buffer, len(read_buffer), byref(bytes_read)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes(read_buffer[: bytes_read.value])


def write_ex(
    ft_handle: I2cMasterHandle,
    dev_address: int,
    flag: TransactionFlag,
    write_data: bytes,
) -> int:
    """Write data to a specified I2C slave device with the specified I2C condition.

    NOTE: This function is supported by the rev. B FT4222H or later!

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Master mode
        dev_address:    Address of target I2C slave device
        flag:           I2C transaction condition flag
        write_data:     Non-empty list of bytes to write

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:            Number of bytes written
    """
    assert (
        0 <= dev_address < (2 ** 16)
    ), "Device address must be an 16b unsigned integer (range 0 - 65 535)"
    assert (
        0 < len(write_data) < (2 ** 16)
    ), "Data to be written must be non-empty and contain less than 2^16 bytes"

    bytes_written = c_uint16()

    result: Ft4222Status = _write_ex(
        ft_handle, dev_address, flag, write_data, len(write_data), byref(bytes_written)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes_written.value


def reset(ft_handle: I2cMasterHandle) -> None:
    """Reset the I2C master device.

    If the I2C bus encounters errors or works abnormally, this function will reset the I2C device.
    It is not necessary to call 'I2C.Master.Init()' again after calling this reset function.
    This function will maintain the original I2C master setting and clear all cache in the device.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Master mode

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _reset(ft_handle)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def get_status(ft_handle: I2cMasterHandle) -> CtrlStatus:
    """Read the status of the I2C master controller.

    This can be used to poll a slave after I2C transmission is complete.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in I2C Master mode

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        Status:  Controller status
    """
    status = c_uint8()

    result: Ft4222Status = _get_status(ft_handle, byref(status))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return CtrlStatus(status.value)


def reset_bus(ft_handle: I2cMasterHandle) -> None:
    """Reset I2C bus.

    If the data line (SDA) is stuck LOW by the slave device, this function will make the master send nine SCK clocks to recover the I2C bus.
    The slave device that held the data line (SDA) LOW will release it within these nine clocks.
    If not, then use the HW reset or cycle power to clear the bus.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Master mode

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _reset_bus(ft_handle)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)



from . import OS_TYPE

import sys
import platform

from ctypes import CDLL, cdll
from typing import Final, Mapping, Optional, Tuple

if sys.version_info < (3, 9):
    import importlib_resources as res
else:
    import importlib.resources as res


ftlib: CDLL
d2lib: CDLL

_COMMON_DLL_PREFIX: str = "pyft4222.wrapper.dll."

_DLL_IMPORT_MAP: Final[Mapping[Tuple[str, str], Tuple[str, str]]] = {
    # Windows
    ("Windows", "AMD64"): (_COMMON_DLL_PREFIX + "win.amd64", "LibFT4222-64.dll"),
    # Linux
    ("Linux", "x86_64"): (_COMMON_DLL_PREFIX + "linux.amd64", "libft4222.so"),
    ("Linux", "aarch64"): (_COMMON_DLL_PREFIX + "linux.aarch64", "libft4222.so"),
    # macOs
    # TODO: Implement!
}

_dll_path = _DLL_IMPORT_MAP.get((OS_TYPE, platform.machine()))


def _get_ft4222_lib() -> CDLL:
    if _dll_path is not None:
        try:
            path = res.files(_dll_path[0]).joinpath(_dll_path[1])
            with res.as_file(path) as fjel:
                return cdll.LoadLibrary(str(fjel))
        except OSError as e:
            print("Unable to load shared library!", file=sys.stderr)
            print(e)
            sys.exit(1)
    else:
        raise RuntimeError("Unsupported OS/CPU combination!")


def _get_d2xx_lib() -> Optional[CDLL]:
    if _dll_path is not None:
        if OS_TYPE == "Windows":
            try:
                return cdll.LoadLibrary("ftd2xx.dll")
            except OSError as e:
                print(
                    "Unable to load D2XX shared library (required on Windows)!",
                    file=sys.stderr,
                )
                print(e)
                sys.exit(1)
        else:
            return None


def init_libraries() -> None:
    global ftlib
    global d2lib

    if "ftlib" not in globals():
        ftlib = _get_ft4222_lib()

    if "d2lib" not in globals():
        temp = _get_d2xx_lib()
        if temp is not None:
            d2lib = temp
        else:
            d2lib = ftlib


init_libraries()



from typing import Generic, List

from pyft4222.handle import GenericProtocolHandle, StreamHandleType
from pyft4222.wrapper import Ft4222Exception, Ft4222Status, GpioTrigger
from pyft4222.wrapper.gpio import (
    GpioHandle,
    PortId,
    get_trigger_status,
    read,
    read_trigger_queue,
    set_input_trigger,
    set_waveform_mode,
    write,
)


class Gpio(
    Generic[StreamHandleType],
    GenericProtocolHandle[GpioHandle, "Gpio", StreamHandleType],
):
    """A class encapsulating GPIO functions."""

    def __init__(self, ft_handle: GpioHandle, stream_handle: StreamHandleType):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:      FT4222 handle initialized in any SPI Master mode
            stream_handle:  Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)

    def read(self, port_id: PortId) -> bool:
        """Read state of the given GPIO port.

        Args:
            port_id:            ID of the port to read

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            bool:               Port state (True -> '1', False -> '0')
        """
        if self._handle is not None:
            return read(self._handle, port_id)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "GPIO has been uninitialized!"
            )

    def write(self, port_id: PortId, state: bool) -> None:
        """Set state of the given GPIO port.

        Args:
            port_id:    ID of the port to set
            state:      New port state (True -> '1', False -> '0')

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            write(self._handle, port_id, state)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "GPIO has been uninitialized!"
            )

    def set_input_trigger(self, port_id: PortId, triggers: GpioTrigger) -> None:
        """Set software trigger conditions for the selected GPIO port.

        Args:
            port_id:        GPIO port ID
            triggers:       OR-map of triggers to enable

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_input_trigger(self._handle, port_id, triggers)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "GPIO has been uninitialized!"
            )

    def get_queued_trigger_event_count(self, port_id: PortId) -> int:
        """Get the size of trigger event queue.

        Args:
            port_id:            GPIO port ID

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            int:                Number of pending events in the event queue
        """
        if self._handle is not None:
            return get_trigger_status(self._handle, port_id)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "GPIO has been uninitialized!"
            )

    def read_trigger_queue(
        self, port_id: PortId, event_read_count: int
    ) -> List[GpioTrigger]:
        """Read events from the trigger event queue.

        Args:
            port_id:            GPIO port ID
            event_read_count:   Number of events to read from queue

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            List[GpioTrigger]:  List of trigger events
        """
        if self._handle is not None:
            return read_trigger_queue(self._handle, port_id, event_read_count)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "GPIO has been uninitialized!"
            )

    def set_waveform_mode(self, enable: bool) -> None:
        """Enable or disable the waveform mode.

        TODO: Add description

        Args:
            enable:             Enable waveform mode?

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_waveform_mode(self._handle, enable)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "GPIO has been uninitialized!"
            )



from ctypes import Structure, POINTER, byref
from ctypes import c_uint, c_void_p, c_bool, c_uint16
from typing import NamedTuple, Union
from enum import IntEnum, auto

from . import Ft4222Status, Ft4222Exception, GpioTrigger, FtHandle
from .dll_loader import ftlib
from .i2c import I2cHandle
from .spi.common import SpiHandle
from .gpio import GpioHandle


# Internal data types


class _RawVersion(Structure):
    _fields_ = [("chip_version", c_uint), ("dll_version", c_uint)]


# External data types


class ClockRate(IntEnum):
    """Enum representing system clock rate.

    All clocks are in MHz.
    """

    SYS_CLK_60 = 0
    """60 MHz"""
    SYS_CLK_24 = auto()
    """24 MHz"""
    SYS_CLK_48 = auto()
    """48 MHz"""
    SYS_CLK_80 = auto()
    """80 MHz"""


class ChipVersion(IntEnum):
    """Enum representing a chip revision.

    Rev. D is newest.
    Rev. A is oldest.

    Note:
        For list of errata, see:
        https://ftdichip.com/wp-content/uploads/2020/08/TN_161_FT4222H-Errata-Technical-Note.pdf
    """

    REV_A = 0x42220100
    REV_B = 0x42220200
    REV_C = 0x42220300
    REV_D = 0x42220400


class SwChipVersion(NamedTuple):
    """NamedTuple representing a combination of library and chip versions."""

    chip_version: ChipVersion
    dll_version: int

    @classmethod
    def from_raw(cls, raw_struct: _RawVersion) -> "SwChipVersion":
        return cls(ChipVersion(raw_struct.chip_version), raw_struct.dll_version)


InitializedHandle = Union[GpioHandle, SpiHandle, I2cHandle]

# DLL functions definitions


_uninitialize = ftlib.FT4222_UnInitialize
_uninitialize.argtypes = [c_void_p]
_uninitialize.restype = Ft4222Status

_set_clock = ftlib.FT4222_SetClock
_set_clock.argtypes = [c_void_p, c_uint]
_set_clock.restype = Ft4222Status

_get_clock = ftlib.FT4222_GetClock
_get_clock.argtypes = [c_void_p, POINTER(c_uint)]
_get_clock.restype = Ft4222Status

_set_wakeup_interrupt = ftlib.FT4222_SetWakeUpInterrupt
_set_wakeup_interrupt.argtypes = [c_void_p, c_bool]
_set_wakeup_interrupt.restype = Ft4222Status

_set_interrupt_trigger = ftlib.FT4222_SetInterruptTrigger
_set_interrupt_trigger.argtypes = [c_void_p, c_uint]
_set_interrupt_trigger.restype = Ft4222Status

_set_suspend_out = ftlib.FT4222_SetSuspendOut
_set_suspend_out.argtypes = [c_void_p, c_bool]
_set_suspend_out.restype = Ft4222Status

_get_max_transfer_size = ftlib.FT4222_GetMaxTransferSize
_get_max_transfer_size.argtypes = [c_void_p, POINTER(c_uint16)]
_get_max_transfer_size.restype = Ft4222Status

_get_version = ftlib.FT4222_GetVersion
_get_version.argtypes = [c_void_p, POINTER(_RawVersion)]
_get_version.restype = Ft4222Status

_chip_reset = ftlib.FT4222_ChipReset
_chip_reset.argtypes = [c_void_p]
_chip_reset.restype = Ft4222Status


def uninitialize(ft_handle: InitializedHandle) -> FtHandle:
    """Release allocated resources.

    This function should be called before calling 'close()'.

    Args:
        ft_handle:  Handle to an initialized FT4222 device

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _uninitialize(ft_handle)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return FtHandle(ft_handle)


def set_clock(ft_handle: FtHandle, clk_rate: ClockRate) -> None:
    """Set the system clock rate.

    The FT4222H supports 4 clock rates: 80MHz, 60MHz, 48MHz, or 24MHz.
    By default, the FT4222H runs at 60MHz clock rate.

    Args:
        ft_handle:  Handle to an opened FT4222 device
        clk_rate:   Desired system clock rate

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_clock(ft_handle, clk_rate)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def get_clock(ft_handle: FtHandle) -> ClockRate:
    """Get the current system clock rate.

    Args:
        ft_handle:  Handle to an opened FT4222 device

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        ClockRate:  Current system clock rate
    """
    clk_rate = c_uint()
    result: Ft4222Status = _get_clock(ft_handle, byref(clk_rate))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return ClockRate(clk_rate.value)


def set_wakeup_interrupt(ft_handle: FtHandle, enable: bool) -> None:
    """Enable or disable wakeup/interrupt.

    Note:
        By default, wake-up/interrupt function is on.

    When Wake up/Interrupt function is on, GPIO3 pin acts as an input pin for wakeup/interrupt.
    While system is in normal mode, GPIO3 acts as an interrupt pin.
    While system is in suspend mode, GPIO3 acts as a wakeup pin.

    Args:
        ft_handle:  Handle to an opened FT4222 device
        enable:     Enable wakeup/interrupt function?

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_wakeup_interrupt(ft_handle, enable)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def set_interrupt_trigger(ft_handle: FtHandle, trigger: GpioTrigger) -> None:
    """Set trigger condition for the pin wakeup/interrupt.

    Note:
        By default, the trigger condition is 'GpioTrigger.RISING'.

    This function configures trigger condition for wakeup/interrupt.

    When GPIO3 acts as wakeup pin, the ft4222H device has the capability to wake up the host.
    Only 'GpioTrigger.RISING' and 'GpioTrigger.FALLING' are valid when GPIO3 act as a wakeup pin.
    It is not necessary to call 'GPIO.init()' function to set up wake-up function.

    When GPIO3 acts as interrupt pin, all trigger conditions can be set.
    The result of trigger status can be inquired by 'GPIO.read_trigger_queue()' or 'GPIO.read()' functions.
    This is because the trigger status is provided by the GPIO pipe.
    Therefore it is necessary to call 'GPIO.init()' function to set up interrupt function.

    For GPIO triggering conditions: 'GpioTrigger.LEVEL_HIGH' and 'GpioTrigger.LEVEL_LOW',
    that can be configured when GPIO3 behaves as an interrupt pin, when the system enters suspend mode,
    these two configurations will act as 'GpioTrigger.RISING' and 'GpioTrigger.FALLING' respectively.

    Args:
        ft_handle:  Handle to an opened FT4222 device
        trigger:    WakeUp/Interrupt condition to set

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_interrupt_trigger(ft_handle, trigger.value)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def set_suspend_out(ft_handle: FtHandle, enable: bool) -> None:
    """Enable or disable suspend out, which will emit a signal when FT4222H enters suspend mode.

    Please note that the suspend-out pin is not available under mode 2.
    By default, suspend-out function is on.
    GPIO2 will be used as an GPIO in output mode for emitting a suspend-out signal.

    When suspend-out function is on, suspend-out pin emits signal according to suspend-out polarity.
    The default value of suspend-out polarity is active high.
    It means suspend-out pin output low in normal mode and output high in suspend mode.
    Suspend-out polarity only can be adjusted by FT_PROG.

    Args:
        ft_handle:  Handle to an opened FT4222 device
        enable:     Enable suspend-out signaling?

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_suspend_out(ft_handle, enable)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def get_max_transfer_size(ft_handle: InitializedHandle) -> int:
    """Get the maximum packet size in a transaction.

    It will be affected by different bus speeds, chip modes, and functions.
    The maximum transfer size is maximum size in writing path.

    Args:
        ft_handle:  Handle to an opened FT4222 device

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:    Maximum packet size in a transaction
    """
    max_size = c_uint16()
    result: Ft4222Status = _get_max_transfer_size(ft_handle, byref(max_size))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return max_size.value


def get_version(ft_handle: FtHandle) -> SwChipVersion:
    """Get the version of FT4222H chip and LibFT4222 library.

    Args:
        ft_handle:  Handle to an opened FT4222 device

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        Version:    NamedTuple containing version information
    """
    version_struct = _RawVersion()
    result: Ft4222Status = _get_version(ft_handle, byref(version_struct))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return SwChipVersion.from_raw(version_struct)


def chip_reset(ft_handle: FtHandle) -> None:
    """Software reset for a device.

    This function is used to attempt to recover system after a failure.
    It is a software reset for device.

    Args:
        ft_handle:  Handle to an opened FT4222 device

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _chip_reset(ft_handle)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)



from abc import ABC
from ctypes import c_void_p
from enum import Enum, auto
from typing import Any, Generic, Literal, TypeVar, Union

from pyft4222.handle import GenericProtocolHandle, StreamHandleType
from pyft4222.wrapper import Ft4222Exception, Ft4222Status
from pyft4222.wrapper.spi import ClkPhase, ClkPolarity, DriveStrength
from pyft4222.wrapper.spi.common import (
    TransactionIdx,
    reset,
    reset_transaction,
    set_driving_strength,
)
from pyft4222.wrapper.spi.slave import (
    EventType,
    SpiSlaveHandle,
    SpiSlaveProtoHandle,
    SpiSlaveRawHandle,
    get_rx_status,
    read,
    set_mode,
    write,
)

SpiSlaveHandleType = TypeVar("SpiSlaveHandleType", bound=SpiSlaveHandle)
SpiSlaveType = TypeVar("SpiSlaveType", "SpiSlaveRaw[Any]", "SpiSlaveProto[Any]")


class SpiModeTag(Enum):
    RAW = auto()
    PROTO = auto()


class SpiSlaveCommon(
    Generic[SpiSlaveHandleType, SpiSlaveType, StreamHandleType],
    GenericProtocolHandle[SpiSlaveHandleType, SpiSlaveType, StreamHandleType],
    ABC,
):
    """A class encapsulating functions common to all SPI Slave modes."""

    def __init__(self, ft_handle: SpiSlaveHandleType, stream_handle: StreamHandleType):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:      FT4222 handle initialized in any SPI Slave mode
            stream_handle:  Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)

    # FIXME: De-duplicate with SPI Master
    def reset_bus(self) -> None:
        """Reset the SPI bus.

        It is not necessary to initialize the SPI Master again.
        The chip retains all settings.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            reset(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Slave has been uninitialized!"
            )

    # FIXME: De-duplicate with SPI Master
    def reset_transaction(self, transaction_idx: TransactionIdx) -> None:
        """Purge transmit and received buffers, reset transaction state.

        Args:
            transaction_idx:    Index of SPI transaction (0 to 3),
            depending on the mode of the chip.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            reset_transaction(self._handle, transaction_idx)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Slave has been uninitialized!"
            )

    # FIXME: De-duplicate with SPI Master
    def set_driving_strength(
        self,
        clk_strength: DriveStrength,
        io_strength: DriveStrength,
        sso_strength: DriveStrength,
    ) -> None:
        """Set driving strength of clk, io and sso pins.

        Note:
            Default driving strength is 4mA.

        Args:
            clk_strength:   Driving strength of the clk pin
            io_strength:    Driving strength of the I/O pins (MISO, MOSI, IO2, IO3)
            sso_strength:   Driving strength of the slave select pins (SSO0, SSO1, SSO2, SSO3)

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_driving_strength(self._handle, clk_strength, io_strength, sso_strength)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Slave has been uninitialized!"
            )

    def set_mode(self, clk_polarity: ClkPolarity, clk_phase: ClkPhase) -> None:
        """Set SPI clock polarity and phase.

        Args:
            clk_polarity:       Clock polarity
            clk_phase:          Clock phase

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_mode(self._handle, clk_polarity, clk_phase)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Slave has been uninitialized!"
            )

    def get_rx_status(self) -> int:
        """Get number of bytes in the Rx queue.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            int:                Number of bytes in the Rx queue
        """
        if self._handle is not None:
            return get_rx_status(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Slave has been uninitialized!"
            )

    def read(self, read_byte_count: int) -> bytes:
        """Read data from the Rx queue.

        Note:
            The FT4222 library adds a zero byte at the beginning.
            FIXME: Is this correct?

        Args:
            read_byte_count:    Number of bytes to read, length <1, 65_535>

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            bytes:              Read data
        """
        if self._handle is not None:
            if 0 < read_byte_count < (2 ** 16):
                return read(self._handle, read_byte_count)
            else:
                raise ValueError("read_byte_count must be in range <1, 65_535>.")
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Slave has been uninitialized!"
            )

    def write(self, write_data: bytes) -> int:
        """Write data into Tx queue.

        Args:
            write_data:     Non-empty list of bytes to write, length <1, 65_535>

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            int:            Number of bytes written
        """
        if self._handle is not None:
            if 0 < len(write_data) < (2 ** 16):
                return write(self._handle, write_data)
            else:
                raise ValueError("write_data length must be in range <1, 65_535>.")
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Slave has been uninitialized!"
            )


class SpiSlaveProto(
    Generic[StreamHandleType],
    SpiSlaveCommon[SpiSlaveProtoHandle, "SpiSlaveProto", StreamHandleType],
):
    """A class encapsulating the SPI Slave protocol mode.

    Attributes:
        tag:    Can be used to disambiguate between SPI Slave in raw/protocol mode
    """

    tag: Literal[SpiModeTag.PROTO]

    def __init__(self, ft_handle: SpiSlaveProtoHandle, stream_handle: StreamHandleType):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:  FT4222 handle initialized in protocol SPI Slave mode
            mode_handle:    Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)
        self.tag = SpiModeTag.PROTO

    # FIXME: Proper typing
    def set_event_notification(self, mask: EventType, param: c_void_p) -> None:
        raise NotImplementedError(
            "Unfortunately this function is yet to be implemented!"
        )
        """TODO: Implement and document"""
        pass


class SpiSlaveRaw(
    Generic[StreamHandleType],
    SpiSlaveCommon[SpiSlaveRawHandle, "SpiSlaveRaw", StreamHandleType],
):
    """A class encapsulating the SPI Slave in raw mode.

    Attributes:
        tag:    Can be used to disambiguate between SPI Slave in raw/protocol mode
    """

    tag: Literal[SpiModeTag.RAW]

    def __init__(self, ft_handle: SpiSlaveRawHandle, stream_handle: StreamHandleType):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:      FT4222 handle initialized in raw SPI Slave mode
            stream_handle:  Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)
        self.tag = SpiModeTag.RAW


SpiSlave = Union[SpiSlaveProto[StreamHandleType], SpiSlaveRaw[StreamHandleType]]



from abc import ABC
from enum import Enum, auto
from typing import Any, Generic, Literal, Optional, TypeVar, Union, overload

from pyft4222.handle import GenericProtocolHandle, StreamHandleType
from pyft4222.wrapper import Ft4222Exception, Ft4222Status
from pyft4222.wrapper.spi import DriveStrength
from pyft4222.wrapper.spi.common import (
    TransactionIdx,
    reset,
    reset_transaction,
    set_driving_strength,
)
from pyft4222.wrapper.spi.master import (
    CsPolarity,
    IoMode,
    SpiMasterHandle,
    SpiMasterMultiHandle,
    SpiMasterSingleHandle,
    multi_read_write,
    set_cs_polarity,
    single_read,
    single_read_write,
    single_write,
)

SpiMasterHandleType = TypeVar("SpiMasterHandleType", bound=SpiMasterHandle)
SpiMasterType = TypeVar("SpiMasterType", bound="SpiMasterCommon[Any, Any, Any]")


class SpiModeTag(Enum):
    SINGLE = auto()
    MULTI = auto()


class SpiMasterCommon(
    Generic[SpiMasterHandleType, SpiMasterType, StreamHandleType],
    GenericProtocolHandle[SpiMasterHandleType, SpiMasterType, StreamHandleType],
    ABC,
):
    """An abstract class encapsulating functions
    common to all SPI Master modes.
    """

    def __init__(self, ft_handle: SpiMasterHandleType, stream_handle: StreamHandleType):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:      FT4222 handle initialized in any SPI Master mode
            stream_handle:  Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)

    def reset_bus(self) -> None:
        """Reset the SPI bus.

        It is not necessary to initialize the SPI Master again.
        The chip retains all settings.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            reset(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )

    def reset_transaction(self, transaction_idx: TransactionIdx) -> None:
        """Purge transmit and receive buffers, reset transaction state.

        Args:
            transaction_idx:    Index of SPI transaction (0 to 3),
            depending on the mode of the chip.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            reset_transaction(self._handle, transaction_idx)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )

    def set_driving_strength(
        self,
        clk_strength: DriveStrength,
        io_strength: DriveStrength,
        sso_strength: DriveStrength,
    ) -> None:
        """Set driving strength of clk, io and sso pins.

        Note:
            Default driving strength is 4mA.

        Args:
            clk_strength:   Driving strength of the clk pin
            io_strength:    Driving strength of the I/O pins (MISO, MOSI, IO2, IO3)
            sso_strength:   Driving strength of the slave select pins (SSO0, SSO1, SSO2, SSO3)

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_driving_strength(self._handle, clk_strength, io_strength, sso_strength)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )

    def set_cs_polarity(self, cs_polarity: CsPolarity) -> None:
        """Set chip select signal polarity.

        Args:
            cs_polarity:        Chip select signal polarity

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_cs_polarity(self._handle, cs_polarity)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )

    @overload
    def set_io_mode(
        self, io_mode: Literal[IoMode.SINGLE]
    ) -> "SpiMasterSingle[StreamHandleType]":
        ...

    @overload
    def set_io_mode(
        self, io_mode: Literal[IoMode.DUAL, IoMode.QUAD]
    ) -> "SpiMasterMulti[StreamHandleType]":
        ...

    @overload
    def set_io_mode(
        self, io_mode: IoMode
    ) -> Union["SpiMasterSingle[StreamHandleType]", "SpiMasterMulti[StreamHandleType]"]:
        ...

    def set_io_mode(
        self, io_mode: IoMode
    ) -> Union["SpiMasterSingle[StreamHandleType]", "SpiMasterMulti[StreamHandleType]"]:
        """Set I/O mode of the SPI Master (i.e., single, dual, quad)

        Note:
            Dual and quad modes are half-duplex (duh).

        Args:
            io_mode:            I/O mode to be set

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMaster:          A class encapsulating the selected mode
        """
        if self._handle is not None:
            temp_handle = self._handle
            self._handle = None

            if io_mode == IoMode.SINGLE:
                return SpiMasterSingle(
                    SpiMasterSingleHandle(temp_handle), self._stream_handle
                )
            else:
                return SpiMasterMulti(
                    SpiMasterMultiHandle(temp_handle), self._stream_handle
                )
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )


class SpiMasterSingle(
    Generic[StreamHandleType],
    SpiMasterCommon[SpiMasterSingleHandle, "SpiMasterSingle", StreamHandleType],
):
    """A class encapsulating the SPI Master in single I/O (full-duplex) mode.

    Attributes:
        tag:    Can be used to disambiguate between SPI Master in single/multi IO mode
    """

    tag: Literal[SpiModeTag.SINGLE]

    def __init__(
        self, ft_handle: SpiMasterSingleHandle, stream_handle: StreamHandleType
    ):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:      FT4222 handle initialized in single SPI Master mode
            stream_handle:  Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)
        self.tag = SpiModeTag.SINGLE

    def single_read(self, read_byte_count: int, end_transaction: bool = True) -> bytes:
        """Read data from an SPI slave.

        Args:
            read_byte_count:    Non-zero number of bytes to read (1 - 65_535)
            end_transaction:    De-assert chip select after a read?

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            bytes:              Read data
        """
        if self._handle is not None:
            result: bytes = bytes()

            while read_byte_count > 0:
                read_len: int = min(read_byte_count, (2 ** 16) - 1)
                trans_end: bool = end_transaction if read_len < (2 ** 16) else False
                result += single_read(self._handle, read_len, trans_end)

                read_byte_count -= read_len

            return result
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )

    def single_write(self, write_data: bytes, end_transaction: bool = True) -> int:
        """Write data to an SPI slave.

        NOTE:
            Data larger than 2^16 - 1 bytes will be done using multiple driver calls.

        Args:
            write_data:         Non-empty list of data to write
            end_transaction:    De-assert slave select after a write?

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            int:                Number of bytes written
        """
        if self._handle is not None:
            write_len: int = len(write_data)
            offset: int = 0
            result: int = 0

            while write_len > 0:
                write_amount: int = min(write_len, (2 ** 16) - 1)
                trans_end: bool = end_transaction if write_len < (2 ** 16) else False
                result += single_write(
                    self._handle, write_data[offset : offset + write_amount], trans_end
                )

                write_len -= write_amount
                offset += write_amount

            return result
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )

    def single_read_write(
        self, write_data: bytes, end_transaction: bool = True
    ) -> bytes:
        """Write and read data concurrently (i.e., full-duplex) from an SPI slave.

        NOTE:
            Data larger than 2^16 - 1 bytes will be done using multiple driver calls.

        Args:
            write_data:         Non-empty list of data to write, length
            end_transaction:    De-assert slave select after a write?

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            bytes:              Read data
        """
        if self._handle is not None:
            write_len: int = len(write_data)
            offset: int = 0
            result: bytes = bytes()

            while write_len > 0:
                write_amount: int = min(write_len, (2 ** 16) - 1)
                trans_end: bool = end_transaction if write_len < (2 ** 16) else False
                result += single_read_write(
                    self._handle, write_data[offset : offset + write_amount], trans_end
                )

                write_len -= write_amount
                offset += write_amount

            return result
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )


class SpiMasterMulti(
    Generic[StreamHandleType],
    SpiMasterCommon[SpiMasterMultiHandle, "SpiMasterMulti", StreamHandleType],
):
    """A class encapsulating the SPI Master in dual or quad I/O (half-duplex) mode.

    Attributes:
        tag:    Can be used to disambiguate between SPI Master in single/multi IO mode
    """

    tag: Literal[SpiModeTag.MULTI]

    def __init__(
        self, ft_handle: SpiMasterMultiHandle, stream_handle: StreamHandleType
    ):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:  FT4222 handle initialized in dual or quad SPI Master mode
            stream_handle:  Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)
        self.tag = SpiModeTag.MULTI

    def multi_read_write(
        self,
        single_write_data: Optional[bytes] = None,
        multi_write_data: Optional[bytes] = None,
        multi_read_byte_count: int = 0,
    ) -> bytes:
        """Write and read data from an SPI slave.

        Note:
            The data are written and read in 3 phases.

        First, the given number of bytes is written using single I/O line. This
        is used mostly for commands (e.g., for EEPROMs, flash memories, etc.).

        Second, the given number of bytes is written using multiple I/O lines.

        Third, the given number of bytes is read using multiple I/O lines.

        Number of bytes to write can be zero.
        Number of bytes to read can be zero.
        But both cannot be zero at the same time.

        The total number of bytes to write cannot be larger than length of
        given data.

        Args:
            single_write_data:          Data to write using single I/O line,
                length <0, 15>              (1st phase)
            multi_write_data:           Data to write using multiple I/O lines,
                length <0, 65_535>          (2nd phase)
            multi_read_byte_count:      Number of bytes to read using multi I/O lines
                length <0, 65_535>          (3rd phase)

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            bytes:              Read data
        """
        single_write_bytes: bytes = (
            single_write_data if single_write_data is not None else bytes()
        )
        multi_write_bytes: bytes = (
            multi_write_data if multi_write_data is not None else bytes()
        )
        write_data_cat: bytes = single_write_bytes + multi_write_bytes

        if self._handle is not None:
            if not (0 <= len(single_write_bytes) < (2 ** 4)):
                raise ValueError("single_write_bytes must be in range <0, 15>.")
            if not (0 <= len(multi_write_bytes) < (2 ** 16)):
                raise ValueError("multi_write_bytes must be in range <0, 65_535>.")
            if not (0 <= multi_read_byte_count < (2 ** 16)):
                raise ValueError("multi_read_bytes must be in range <0, 65_535>.")
            if (
                len(single_write_bytes) + len(multi_write_bytes) + multi_read_byte_count
            ) <= 0:
                raise ValueError(
                    "Total number of bytes to read and write must be non-zero."
                )

            return multi_read_write(
                self._handle,
                write_data_cat,
                len(single_write_bytes),
                len(multi_write_bytes),
                multi_read_byte_count,
            )
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "SPI Master has been uninitialized!"
            )


SpiMaster = Union[SpiMasterSingle[StreamHandleType], SpiMasterMulti[StreamHandleType]]



from ctypes import (
    POINTER,
    byref,
    c_bool,
    c_char_p,
    c_uint,
    c_uint8,
    c_uint16,
    c_uint32,
    c_void_p,
)
from enum import IntEnum, IntFlag, auto
from typing import Literal, NewType, Optional, Union, overload

from koda import Err, Ok, Result

from .. import Ft4222Exception, Ft4222Status, FtHandle
from ..dll_loader import ftlib
from . import ClkPhase, ClkPolarity

SpiMasterSingleHandle = NewType("SpiMasterSingleHandle", FtHandle)
SpiMasterMultiHandle = NewType("SpiMasterMultiHandle", FtHandle)
SpiMasterHandle = Union[SpiMasterSingleHandle, SpiMasterMultiHandle]


class IoMode(IntEnum):
    """Enum representing possible SPI data I/O modes/counts."""

    SINGLE = 1
    """Data are transferred full-duplex using MISO/MOSI lines."""
    DUAL = 2
    """Data are transferred half-duplex using MISO, MOSI lines (bit0, bit1)."""
    QUAD = 4
    """Data are transferred half-duples using
    MISO, MOSI, IO2, IO3 lines (bit0, bit1, bit2, bit3).
    """


class ClkDiv(IntEnum):
    """Enum representing possible clock divisors in SPI Master mode."""

    CLK_NONE = 0
    CLK_DIV_2 = auto()  # 1/2   System Clock
    CLK_DIV_4 = auto()  # 1/4   System Clock
    CLK_DIV_8 = auto()  # 1/8   System Clock
    CLK_DIV_16 = auto()  # 1/16  System Clock
    CLK_DIV_32 = auto()  # 1/32  System Clock
    CLK_DIV_64 = auto()  # 1/64  System Clock
    CLK_DIV_128 = auto()  # 1/128 System Clock
    CLK_DIV_256 = auto()  # 1/256 System Clock
    CLK_DIV_512 = auto()  # 1/512 System Clock


class CsPolarity(IntEnum):
    """Enum representing possible chip_select signal polarities."""

    ACTIVE_LOW = 0
    """Active-low chip select (DEFAULT)."""
    ACTIVE_HIGH = auto()
    """Active-high chip select."""


class SsoMap(IntFlag):
    """Enum representing a slave select mapping flag."""

    SS_0 = 1
    SS_1 = 2
    SS_2 = 4
    SS_3 = 8


_init = ftlib.FT4222_SPIMaster_Init
_init.argtypes = [c_void_p, c_uint, c_uint, c_uint, c_uint, c_uint8]
_init.restype = Ft4222Status

_set_cs = ftlib.FT4222_SPIMaster_SetCS
_set_cs.argtypes = [c_void_p, c_uint]
_set_cs.restype = Ft4222Status

_set_lines = ftlib.FT4222_SPIMaster_SetLines
_set_lines.argtypes = [c_void_p, c_uint]
_set_lines.restype = Ft4222Status

_single_read = ftlib.FT4222_SPIMaster_SingleRead
_single_read.argtypes = [
    c_void_p,
    POINTER(c_uint8),
    c_uint16,
    POINTER(c_uint16),
    c_bool,
]
_single_read.restype = Ft4222Status

_single_write = ftlib.FT4222_SPIMaster_SingleWrite
_single_write.argtypes = [c_void_p, c_char_p, c_uint16, POINTER(c_uint16), c_bool]
_single_write.restype = Ft4222Status

_single_read_write = ftlib.FT4222_SPIMaster_SingleReadWrite
_single_read_write.argtypes = [
    c_void_p,
    POINTER(c_uint8),
    c_char_p,
    c_uint16,
    POINTER(c_uint16),
    c_bool,
]
_single_read_write.restype = Ft4222Status

_multi_read_write = ftlib.FT4222_SPIMaster_MultiReadWrite
_multi_read_write.argtypes = [
    c_void_p,
    POINTER(c_uint8),
    c_char_p,
    c_uint8,
    c_uint16,
    c_uint16,
    POINTER(c_uint32),
]
_multi_read_write.restype = Ft4222Status


@overload
def init(
    ft_handle: FtHandle,
    io_mode: Literal[IoMode.SINGLE],
    clock_div: ClkDiv,
    clk_polarity: ClkPolarity,
    clk_phase: ClkPhase,
    sso_map: SsoMap,
) -> Result[SpiMasterSingleHandle, Ft4222Status]:
    ...


@overload
def init(
    ft_handle: FtHandle,
    io_mode: Union[Literal[IoMode.DUAL], Literal[IoMode.QUAD]],
    clock_div: ClkDiv,
    clk_polarity: ClkPolarity,
    clk_phase: ClkPhase,
    sso_map: SsoMap,
) -> Result[SpiMasterMultiHandle, Ft4222Status]:
    ...


@overload
def init(
    ft_handle: FtHandle,
    io_mode: IoMode,
    clock_div: ClkDiv,
    clk_polarity: ClkPolarity,
    clk_phase: ClkPhase,
    sso_map: SsoMap,
) -> Union[
    Result[SpiMasterSingleHandle, Ft4222Status],
    Result[SpiMasterMultiHandle, Ft4222Status],
]:
    ...


def init(
    ft_handle: FtHandle,
    io_mode: IoMode,
    clock_div: ClkDiv,
    clk_polarity: ClkPolarity,
    clk_phase: ClkPhase,
    sso_map: SsoMap,
) -> Union[
    Result[SpiMasterSingleHandle, Ft4222Status],
    Result[SpiMasterMultiHandle, Ft4222Status],
]:
    """Initialize the FT4222H as an SPI master.

    Args:
        ft_handle:      Handle to an open FT4222 device
        io_mode:        Data transmission mode (single, dual, quad)
        clock_div:      SPI clock rate division ratio (spi_clock = (sys_clk / clock_div))
        clk_polarity:   Clock polarity (idle low/high)
        clk_phase:      Clock phase (data sampled on the leading [first] or trailing [second] clock edge)
        sso_map:        Slave selection output pin map

    Returns:
        Result:         Handle to initialized SPI Master FT4222 device
    """
    result: Ft4222Status = _init(
        ft_handle, io_mode, clock_div, clk_polarity, clk_phase, sso_map
    )

    if result == Ft4222Status.OK:
        if io_mode == IoMode.SINGLE:
            return Ok(SpiMasterSingleHandle(ft_handle))
        elif io_mode in {IoMode.DUAL, IoMode.QUAD}:
            return Ok(SpiMasterMultiHandle(ft_handle))
        else:
            return Err(result)
    else:
        return Err(result)


def set_cs_polarity(ft_handle: SpiMasterHandle, cs_polarity: CsPolarity) -> None:
    """Change polarity of chip select signal.

    Default chip select is active low.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in SPI Master mode
        cs_polarity:    Polarity of chip select signal

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_cs(ft_handle, cs_polarity)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


@overload
def set_lines(
    ft_handle: SpiMasterHandle, io_mode: Literal[IoMode.SINGLE]
) -> SpiMasterSingleHandle:
    ...


@overload
def set_lines(
    ft_handle: SpiMasterHandle, io_mode: Literal[IoMode.DUAL, IoMode.QUAD]
) -> SpiMasterMultiHandle:
    ...


@overload
def set_lines(ft_handle: SpiMasterHandle, io_mode: IoMode) -> SpiMasterHandle:
    ...


def set_lines(ft_handle: SpiMasterHandle, io_mode: IoMode) -> SpiMasterHandle:
    """Switch the FT4222H SPI Master to single, dual, or quad mode.

    This overrides the mode passed to 'SPI.Master.init()' function.
    This might be needed if a device accepts commands in single mode but data transfer is using dual or quad mode.

    Args:
        ft_handle:  Handle to an initialized FT4222 device in SPI Master mode
        io_mode:    Desired IO mode (single, dual, quad)

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        SpiMasterHandle:    Handle to an initialized FT4222 device in SPI Master mode with selected io setting
    """
    result: Ft4222Status = _set_lines(ft_handle, io_mode)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    if io_mode == IoMode.SINGLE:
        return SpiMasterSingleHandle(ft_handle)
    elif io_mode in [IoMode.DUAL, IoMode.QUAD]:
        return SpiMasterMultiHandle(ft_handle)
    else:
        raise Ft4222Exception(result, "Invalid IoMode! Cannot use 'IoMode.NONE'.")


def single_read(
    ft_handle: SpiMasterSingleHandle, read_byte_count: int, end_transaction: bool = True
) -> bytes:
    """Under SPI single mode, read data from an SPI slave.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Master mode with 'IoMode.IO_SINGLE' setting
        read_byte_count:    Positive number of bytes to read
        end_transaction:    De-assert slave select pin at the end of transaction?

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        bytes:              Read data (length can be lower than requested)
    """
    assert (
        0 < read_byte_count < (2 ** 16)
    ), "Number of bytes to read must be positive and less than 2^16"

    buffer = (c_uint8 * read_byte_count)()
    bytes_transferred = c_uint16()

    result: Ft4222Status = _single_read(
        ft_handle, buffer, read_byte_count, byref(bytes_transferred), end_transaction
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes(buffer)


def single_write(
    ft_handle: SpiMasterSingleHandle, write_data: bytes, end_transaction: bool = True
) -> int:
    """Under SPI single mode, write data to an SPI slave.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Master mode with 'IoMode.IO_SINGLE' setting
        write_data:         Non-empty list of bytes to be written
        end_transaction:    De-assert slave select pin at the end of transaction?

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:                Number of transmitted bytes
    """
    assert (
        0 < len(write_data) < (2 ** 16)
    ), "Data to be written must be non-empty and contain less than 2^16 bytes"

    bytes_transferred = c_uint16()
    result: Ft4222Status = _single_write(
        ft_handle,
        write_data,
        len(write_data),
        byref(bytes_transferred),
        end_transaction,
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes_transferred.value


def single_read_write(
    ft_handle: SpiMasterSingleHandle, write_data: bytes, end_transaction: bool = True
) -> bytes:
    """Under SPI single mode, full-duplex write data to and read data from an SPI slave.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Master mode with 'IoMode.IO_SINGLE' setting
        write_data:         Non-empty list of data to be written
        end_transaction:    De-assert slave select pin at the end of transaction?

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        bytes:              Received data
    """
    assert (
        0 < len(write_data) < (2 ** 16)
    ), "Data to be written must be non-empty and contain less than 2^16 bytes"

    bytes_transferred = c_uint16()
    read_buffer = (c_uint8 * len(write_data))()

    result: Ft4222Status = _single_read_write(
        ft_handle,
        read_buffer,
        write_data,
        len(write_data),
        byref(bytes_transferred),
        end_transaction,
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes(read_buffer)


def multi_read_write(
    ft_handle: SpiMasterMultiHandle,
    write_data: Optional[bytes],
    single_write_byte_count: int,
    multi_write_byte_count: int,
    multi_read_byte_count: int,
) -> bytes:
    """Under SPI dual or quad mode, write data to and read data from an SPI slave.

    It is a mixed protocol initiated with a single write transmission,
    which may be an SPI command and dummy cycles,
    and followed by multi-io-write and multi-io-read transmission that use 2/4 signals in parallel for the data.
    All three parts of the protocol are optional.

    Args:
        ft_handle:                  Handle to an initialized FT4222 device in SPI Master mode with 'IoMode.IO_DUAL' or 'IoMode'IO_QUAD' setting
        write_data:                 Data to be written
        single_write_byte_count:    Number of bytes to be written out using single IO line      (1st phase)
        multi_write_byte_count:     Number of bytes to be written out using multi IO lines      (2nd phase)
        multi_read_byte_count:      Number of bytes to be read using multi IO lines             (3rd phase)

    Raises:
        Ft4222Exception:            In case of unexpected error

    Returns:
        bytes:                      Read data (if any)
    """
    assert (
        0 <= single_write_byte_count < (2 ** 4)
    ), "Number of single-write bytes must be non-negative and less than 16"
    assert (
        0 <= multi_write_byte_count < (2 ** 16)
    ), "Number of multi-write bytes must be non-negative and less than 2^16 (65 536)"
    assert (
        0 <= multi_read_byte_count < (2 ** 16)
    ), "Number of multi-read bytes must be non-negative and less than 2^16 (65 536)"
    assert (
        single_write_byte_count + multi_write_byte_count + multi_read_byte_count
    ) > 0, "Total number of bytes written/read must be non-zero"
    if write_data is None:
        assert (
            single_write_byte_count + multi_write_byte_count
        ) == 0, "Number of bytes to write must be zero in case the write data are None"
    else:
        assert len(write_data) < (
            2 ** 16
        ), "Data to be written must have size less than 2^16 bytes"
        assert (single_write_byte_count + multi_write_byte_count) <= len(
            write_data
        ), "Length of data to write is longer than given data"

    read_buffer = (c_uint8 * multi_read_byte_count)()
    bytes_read = c_uint32()
    result: Ft4222Status = _multi_read_write(
        ft_handle,
        read_buffer,
        write_data,
        single_write_byte_count,
        multi_write_byte_count,
        multi_read_byte_count,
        byref(bytes_read),
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes(read_buffer[: bytes_read.value])



"""TODO: Add module information

"""

from enum import Enum, auto
from typing import Callable, Dict, Final, List, Tuple, Type, Union

from koda import Err, Ok, Result

from pyft4222.stream import GpioStream, ProtocolStream, SpiStream
from pyft4222.wrapper import OS_TYPE, FtHandle
from pyft4222.wrapper import ftd2xx as ftd
from pyft4222.wrapper import gpio as wgpio
from pyft4222.wrapper.common import uninitialize

VERSION: Final = "0.1.1"

_DEFAULT_GPIO_DIRS: Final[wgpio.DirTuple] = (
    wgpio.Direction.INPUT,
    wgpio.Direction.INPUT,
    wgpio.Direction.INPUT,
    wgpio.Direction.INPUT,
)


def _disambiguate_modes(handle: FtHandle) -> Union[GpioStream, SpiStream]:
    """Disambiguate between chip configuration mode 1 and 2.

    Args:
        handle:             Handle to an opened FT4222 interface (in mode_1_2)

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        Union[GpioHandle, SpiMasterHandle]:     Returns disambiguated handle
    """
    result = wgpio.init(handle, _DEFAULT_GPIO_DIRS)
    if isinstance(result, Ok):
        return GpioStream(uninitialize(result.val))
    else:
        return SpiStream(handle)


_Ft4222HandleType = Union[
    Type[ProtocolStream],
    Type[GpioStream],
    Type[SpiStream],
    Callable[[FtHandle], Union[GpioStream, SpiStream]],
]

Ft4222Handle = Union[ProtocolStream, GpioStream, SpiStream]

_MODE_MAP: Final[Dict[Tuple[str, ftd.DeviceType], _Ft4222HandleType]] = {
    # Mode 0
    ("FT4222 A", ftd.DeviceType.DEV_4222H_0): ProtocolStream,
    ("FT4222 B", ftd.DeviceType.DEV_4222H_0): GpioStream,
    # Mode 1 or Mode 2
    ("FT4222 A", ftd.DeviceType.DEV_4222H_1_2): SpiStream,
    ("FT4222 B", ftd.DeviceType.DEV_4222H_1_2): SpiStream,
    ("FT4222 C", ftd.DeviceType.DEV_4222H_1_2): SpiStream,
    ("FT4222 D", ftd.DeviceType.DEV_4222H_1_2): _disambiguate_modes,
    # Mode 3
    ("FT4222", ftd.DeviceType.DEV_4222H_3): ProtocolStream,
}


class FtError(Enum):
    INVALID_MODE = auto()
    INVALID_IDX = auto()
    INVALID_HANDLE = auto()
    UNKNOWN_FAILURE = auto()


def _get_mode_handle(ft_handle: FtHandle) -> Result[Ft4222Handle, FtError]:
    """Resolve and instantiate a correct device configuration class.

    Args:
        ft_handle:  Handle to open FT4222 device stream

    Returns:
        Result[Ft4222Handle, FtError]:
    """
    dev_details = ftd.get_device_info(ft_handle)
    if dev_details is not None:
        handle_type = _MODE_MAP.get((dev_details.description, dev_details.dev_type))
        if handle_type is not None:
            return Ok(handle_type(ft_handle))
        else:
            return Err(FtError.INVALID_MODE)
    else:
        return Err(FtError.INVALID_HANDLE)


def _validate_dev_idx(dev_idx: int) -> bool:
    """Validate given D2XX device index.

    Args:
        dev_idx:    D2XX device index

    Returns:
        bool:       Is given device index valid?
    """
    return 0 <= dev_idx < (2**31)


def get_device_info_list() -> List[ftd.DeviceInfo]:
    """Get list containing information about available (connected) D2XX devices.

    Raises:
        FtException:        In case of unexpected error

    Returns:
        List[DeviceInfo]:   List containing details about connected D2XX devices
    """
    return ftd.get_device_info_list()


def get_device_info_detail(dev_idx: int) -> Result[ftd.DeviceInfo, FtError]:
    """Get information about device at the given index (if any).

    Args:
        dev_idx:        Zero-base index of the device (0 to (2^31 - 1))

    Raises:
        FtException:    In case of unexpected error

    Returns:
        Result[DeviceInfo, FtError]:    Structure containing information
        about the device
    """
    if _validate_dev_idx(dev_idx):
        result = ftd.get_device_info_detail(dev_idx)
        if result is not None:
            return Ok(result)

    return Err(FtError.INVALID_IDX)


def open_by_idx(dev_idx: int) -> Result[Ft4222Handle, FtError]:
    """Open FT4222 device stream using 'device index'.

    The device index is 0-based and can be found by calling
    'get_device_info_list()' for example.

    Args:
        dev_id:         Device index

    Raises:
        FtException:    In case of unexpected error

    Returns:
        Ft4222Handle:   Handle encapsulating modes available in the
        current device configuration mode
    """
    if _validate_dev_idx(dev_idx):
        ft_handle = ftd.open_by_idx(dev_idx)
        if isinstance(ft_handle, Ok):
            return _get_mode_handle(ft_handle.val)
        else:
            return Err(FtError.INVALID_HANDLE)
    else:
        return Err(FtError.INVALID_IDX)


def open_by_serial(serial_num: str) -> Result[Ft4222Handle, FtError]:
    """Open FT4222 device stream using 'device serial number'.

    Args:
        serial_num:     A device serial number string

    Returns:
        Result[Ft4222Handle, FtError]:
    """
    ft_handle = ftd.open_by_serial(serial_num)
    if isinstance(ft_handle, Ok):
        return _get_mode_handle(ft_handle.val)
    else:
        return Err(FtError.INVALID_HANDLE)


def open_by_description(dev_description: str) -> Result[Ft4222Handle, FtError]:
    """Open FT4222 device stream using 'device description'.

    Args:
        dev_description:    Device description string

    Returns:
        Result[Ft4222Handle, FtError]:
    """
    ft_handle = ftd.open_by_description(dev_description)
    if isinstance(ft_handle, Ok):
        return _get_mode_handle(ft_handle.val)
    else:
        return Err(FtError.INVALID_HANDLE)


if OS_TYPE != "Linux":

    def open_by_location(location_id: int) -> Result[Ft4222Handle, FtError]:
        """Open FT4222 device stream using 'device USB location ID'.

        Args:
            location_id:    Device USB location ID

        Returns:
            Result[Ft4222Handle, FtError]:
        """
        ft_handle = ftd.open_by_location(location_id)
        if isinstance(ft_handle, Ok):
            return _get_mode_handle(ft_handle.val)
        else:
            return Err(FtError.INVALID_HANDLE)



from typing import Final, Literal, Set, Union
from ctypes import c_void_p, c_uint8, c_uint

from .slave import SpiSlaveHandle
from .master import SpiMasterHandle

from . import DriveStrength
from ..dll_loader import ftlib
from .. import Ft4222Exception, Ft4222Status

SpiHandle = Union[SpiMasterHandle, SpiSlaveHandle]

TransactionIdx = Literal[0, 1, 2, 3]


_SOFT_ERROR_SET: Final[Set[Ft4222Status]] = {
    Ft4222Status.OK,
    Ft4222Status.INVALID_HANDLE,
    Ft4222Status.DEVICE_NOT_FOUND,
    Ft4222Status.DEVICE_NOT_OPENED,
}


_reset = ftlib.FT4222_SPI_Reset
_reset.argtypes = [c_void_p]
_reset.restype = Ft4222Status

_reset_transaction = ftlib.FT4222_SPI_ResetTransaction
_reset_transaction.argtypes = [c_void_p, c_uint8]
_reset_transaction.restype = Ft4222Status

_set_driving_strength = ftlib.FT4222_SPI_SetDrivingStrength
_set_driving_strength.argtypes = [c_void_p, c_uint, c_uint, c_uint]
_set_driving_strength.restype = Ft4222Status


def reset(ft_handle: SpiHandle) -> None:
    """Reset the SPI master or slave device.

    If the SPI bus encounters errors or works abnormally, this function will reset the SPIdevice.
    It is not necessary to call SPI init function again after calling this reset function.
    It retains all original setting of SPI.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Master/Slave mode

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _reset(ft_handle)

    if result not in _SOFT_ERROR_SET:
        raise Ft4222Exception(result)


def reset_transaction(ft_handle: SpiHandle, spi_idx: TransactionIdx) -> None:
    """Reset the SPI transaction.

    Purge receive and transmit buffers in the device and reset the transaction state.

    Args:
        ft_handle:  Handle to an initialized FT4222 device in SPI Master/Slave mode
        spi_idx:    Index of SPI transaction (0 - 3), depending on the mode of the chip

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    assert 0 <= spi_idx <= 3, "Invalid SPI transaction index (0 - 3)"

    result: Ft4222Status = _reset_transaction(ft_handle, spi_idx)

    if result not in _SOFT_ERROR_SET:
        raise Ft4222Exception(result)


def set_driving_strength(
    ft_handle: SpiHandle,
    clk_strength: DriveStrength,
    io_strength: DriveStrength,
    sso_strength: DriveStrength,
) -> None:
    """Set the driving strength of clk, io, and sso pins.

    Default driving strength of all spi pins is 'SPI.DriveStrength.DS_4MA'.
    Unless there is some hardware wiring requirement for device,
    setting driving strength to 4 mA should be enough.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in SPI Master/Slave mode
        clk_strength:   Driving strength of the clk pin
        io_strength:    Driving strength of the I/O pins (MISO, MOSI, IO2, IO3)
        sso_strength:   Driving strength of the slave select pins (SSO0, SSO1, SSO2, SSO3)

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_driving_strength(
        ft_handle, clk_strength, io_strength, sso_strength
    )

    if result not in _SOFT_ERROR_SET:
        raise Ft4222Exception(result)



from ctypes import (
    POINTER,
    Structure,
    byref,
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_uint32,
    c_void_p,
    create_string_buffer,
)
from enum import IntEnum, IntFlag, auto
from typing import Final, List, NamedTuple, Optional

from koda import Err, Ok, Result

from . import OS_TYPE, FtException, FtHandle, FtStatus
from .dll_loader import d2lib

# DLL function protoypes declaration


_create_device_info_list = d2lib.FT_CreateDeviceInfoList
_create_device_info_list.argtypes = [POINTER(c_uint)]
_create_device_info_list.restype = FtStatus

_get_device_info_list = d2lib.FT_GetDeviceInfoList
_get_device_info_list.argtypes = []
_get_device_info_list.restype = FtStatus

_get_device_info_detail = d2lib.FT_GetDeviceInfoDetail
_get_device_info_detail.argtypes = []
_get_device_info_detail.restype = FtStatus

_open = d2lib.FT_Open
_open.argtypes = [c_int, POINTER(c_void_p)]
_open.restype = FtStatus

_open_ex = d2lib.FT_OpenEx
_open_ex.argtypes = [c_void_p, c_uint, POINTER(c_void_p)]
_open_ex.restype = FtStatus

_close = d2lib.FT_Close
_close.argtypes = [c_void_p]
_close.restype = FtStatus

_get_device_info = d2lib.FT_GetDeviceInfo
_get_device_info.argtypes = [
    c_void_p,
    POINTER(c_uint),
    POINTER(c_uint),
    c_char_p,
    c_char_p,
    c_void_p,
]
_get_device_info.restype = FtStatus

_get_driver_version = d2lib.FT_GetDriverVersion
_get_driver_version.argtypes = [c_void_p, POINTER(c_uint)]
_get_driver_version.restype = FtStatus

_purge = d2lib.FT_Purge
_purge.argtypes = [c_void_p, c_uint]
_purge.restype = FtStatus

_reset_device = d2lib.FT_ResetDevice
_reset_device.argtypes = [c_void_p]
_reset_device.restype = FtStatus

# Constants

# Max. length of C string containing D2XX device serial number,
# according to the D2XX documentation
_SERIAL_NUMBER_MAX_LEN: Final[int] = 16
# Max. length of C string containing D2XX device description,
# according to the D2XX documentation
_DESCRIPTION_MAX_LEN: Final[int] = 64

# Data types


class _RawDeviceInfoListNode(Structure):
    _fields_ = [
        ("flags", c_uint),
        ("type", c_uint),
        ("id", c_uint),
        ("loc_id", c_uint),
        ("serial_number", c_char * _SERIAL_NUMBER_MAX_LEN),
        ("description", c_char * _DESCRIPTION_MAX_LEN),
        ("handle", c_void_p),
    ]


class DeviceFlags(IntFlag):
    """Class representing the D2XX 'device information flags' enum."""

    OPEN = 1
    """Device is currently opened."""
    HIGH_SPEED = 2
    """Device is using USB High-Speed signaling rate."""


class DeviceType(IntEnum):
    """Class representing the 'FT_DEVICE' D2XX enum."""

    DEV_BM = 0
    DEV_AM = auto()
    DEV_100AX = auto()
    DEV_UNKNOWN = auto()
    DEV_2232C = auto()
    DEV_232R = auto()
    DEV_2232H = auto()
    DEV_4232H = auto()
    DEV_232H = auto()
    DEV_X_SERIES = auto()
    DEV_4222H_0 = auto()
    DEV_4222H_1_2 = auto()
    DEV_4222H_3 = auto()
    DEV_4222_PROG = auto()
    DEV_900 = auto()
    DEV_930 = auto()
    DEV_UMFTPD3A = auto()


class DeviceInfo(NamedTuple):
    """NamedTuple encapsulating the 'FT_DEVICE_INFO_LIST_NODE' D2XX data struct."""

    flags: DeviceFlags
    """Special device flags. See 'DeviceFlags'."""
    dev_type: DeviceType
    """Device type. See 'DeviceType'."""
    idx: int
    """Device index in device information list."""
    loc_id: int
    """Device USB Location ID."""
    serial_number: str
    """Serial number of the device."""
    description: str
    """Device description."""
    handle: FtHandle
    """A C pointer to device handle; NULL if not opened"""

    @classmethod
    def from_raw(cls, raw_node: _RawDeviceInfoListNode) -> "DeviceInfo":
        return cls(
            flags=DeviceFlags(raw_node.flags),
            dev_type=DeviceType(raw_node.type),
            idx=raw_node.id,
            loc_id=raw_node.loc_id,
            serial_number=raw_node.serial_number.decode("utf-8"),
            description=raw_node.description.decode("utf-8"),
            handle=FtHandle(raw_node.handle),
        )


class DriverVersion(NamedTuple):
    """NamedTuple encapsulating the driver version information."""

    major: int
    minor: int
    build_version: int


class _OpenExFlag(IntEnum):
    BY_SERIAL_NUMBER = 1
    BY_DESCRIPTION = 2
    BY_LOCATION = 4


class BufferType(IntFlag):
    """Enum for selecting the buffer purge type."""

    RX = 1
    """Purge receive buffers."""
    TX = 2
    """Purge transmit buffers."""


class ShortDeviceInfo(NamedTuple):
    """NamedTuple encapsulating the short device information."""

    dev_type: DeviceType
    """Device type. See 'DeviceType' enum."""
    dev_idx: int
    """Device index in device information list."""
    serial_number: str
    """Device serial number."""
    description: str
    """Device description"""


def create_device_info_list() -> int:
    """Create a device info list.

    This function builds a device information list and returns
    the number of D2XX devices connected to the system.
    The list contains information about both unopened and open devices.

    An application can use this function to get the number
    of devices attached to the system.
    It can then allocate space for the device information list
    and retrieve the list using get_device_info_list.
    If the devices connected to the system change,
    the device info list will not be updated until
    'create_device_info_list()' is called again.

    Raises:
        FtException:    In case of unexpected error

    Returns:
        int:            Number of connected D2XX devices
    """
    dev_count = c_uint()
    result: FtStatus = _create_device_info_list(byref(dev_count))

    if result != FtStatus.OK:
        raise FtException(result)

    return dev_count.value


def get_device_info_list() -> List[DeviceInfo]:
    """Return list containing details about connected D2XX devices.

    This function returns a device information list containing information
    about all connected D2XX devices.

    Args:
        dev_count:          Max. number of devices to construct list for

    Raises:
        FtException:        In case of unexpected error

    Returns:
        List[DeviceInfo]:   List containing details about connected D2XX devices
    """
    dev_count = create_device_info_list()

    array_elem_count = c_uint(dev_count)
    raw_list = (_RawDeviceInfoListNode * dev_count)()

    result: FtStatus = _get_device_info_list(raw_list, byref(array_elem_count))

    if result != FtStatus.OK:
        raise FtException(result)

    return list(map(DeviceInfo.from_raw, raw_list))[: array_elem_count.value]


def get_device_info_detail(dev_idx: int) -> Optional[DeviceInfo]:
    """This function returns an entry from the device information list.

    Args:
        dev_id:         Index of device in device info list (non-negative)

    Raises:
        FtException:    In case of unexpected error

    Returns:
        device details (if id exists)
    """
    assert 0 <= dev_idx <= (2 ** 31) - 1, "Device ID must be a non-negative integer"

    idx = c_uint(dev_idx)
    flags = c_uint()
    dev_type = c_uint()
    idx = c_uint()
    loc_id = c_uint()
    serial_number = create_string_buffer(_SERIAL_NUMBER_MAX_LEN)
    description = create_string_buffer(_DESCRIPTION_MAX_LEN)
    handle = c_void_p()

    result: FtStatus = _get_device_info_detail(
        idx,
        byref(flags),
        byref(dev_type),
        byref(idx),
        byref(loc_id),
        serial_number,
        description,
        byref(handle),
    )

    if result == FtStatus.OK:
        return DeviceInfo(
            flags=DeviceFlags(flags.value),
            dev_type=DeviceType(dev_type.value),
            idx=idx.value,
            loc_id=loc_id.value,
            serial_number=serial_number.value.decode("utf-8"),
            description=description.value.decode("utf-8"),
            handle=FtHandle(handle),
        )
    elif result in {FtStatus.DEVICE_NOT_FOUND, FtStatus.INVALID_HANDLE}:
        return None
    else:
        raise FtException(result)


def open_by_idx(dev_idx: int) -> Result[FtHandle, FtStatus]:
    """Open the device using 'device index' and return a device handle.

    NOTE:
        Although this function can be used to open multiple devices
        by setting iDevice to 0, 1, 2 etc., there is no ability to open
        a specific device.
        To open named devices, use functions: 'open_by_serial()',
        'open_by_description()', 'open_by_location()'.

    Args:
        dev_id:     D2XX device index

    Returns:
        D2XX device handle (optional)
    """
    assert 0 <= dev_idx <= (2 ** 31) - 1, "Device ID must be a non-negative integer"

    ft_handle = c_void_p()
    result: FtStatus = _open(dev_idx, byref(ft_handle))

    if result == FtStatus.OK:
        return Ok(FtHandle(ft_handle))
    else:
        return Err(result)


def open_by_serial(serial_num: str) -> Result[FtHandle, FtStatus]:
    """Open the specified device using 'serial number' and return a device handle.

    The device is specified by its serial number.

    This function can also be used to open multiple devices simultaneously.

    Args:
        serial_num:     Serial number of the D2XX device

    Returns:
        D2XX device handle (optional)
    """
    ft_handle = c_void_p()
    result: FtStatus = _open_ex(
        serial_num.encode("utf-8"), _OpenExFlag.BY_SERIAL_NUMBER, byref(ft_handle)
    )

    if result == FtStatus.OK:
        return Ok(FtHandle(ft_handle))
    else:
        return Err(result)


def open_by_description(dev_description: str) -> Result[FtHandle, FtStatus]:
    """Open the specified device using 'device description' and return a device handle.

    The device is specified by its description string.

    This function can also be used to open multiple devices simultaneously.

    Arguments:
        dev_description:    Description string of the D2XX device

    Return:
        D2XX device handle (optional)
    """
    ft_handle = c_void_p()
    result: FtStatus = _open_ex(
        dev_description.encode("utf-8"), _OpenExFlag.BY_DESCRIPTION, byref(ft_handle)
    )

    if result == FtStatus.OK:
        return Ok(FtHandle(ft_handle))
    else:
        return Err(result)


# This function is not supported on Linux and Windows CE
# REVIEW: Windows CE is unsupported by Python so no need to check?
if OS_TYPE != "Linux":

    def open_by_location(location_id: int) -> Result[FtHandle, FtStatus]:
        """Open the specified device using 'location ID' and return a device handle.

        The device is specified by its location ID.

        This function can also be used to open multiple devices simultaneously.

        Note:
            This method is not supported on Windows CE and Linux.

        Arguments:
            location_id:    Location ID of the D2XX device

        Returns:
            Result[FtHandle, FtStatus]: D2XX device handle (if location exists)
        """
        ft_handle = c_void_p()
        result: FtStatus = _open_ex(
            location_id, _OpenExFlag.BY_LOCATION, byref(ft_handle)
        )

        if result == FtStatus.OK:
            return Ok(FtHandle(ft_handle))
        else:
            return Err(result)


def close_handle(ft_handle: FtHandle) -> None:
    """Close an open device.

    Args:
        ft_handle:      Handle to an opened D2XX device

    Raises:
        FtException:    In case of unexpected error
    """
    result: FtStatus = _close(ft_handle)

    if result not in [
        FtStatus.OK,
        FtStatus.DEVICE_NOT_OPENED,
        FtStatus.INVALID_HANDLE,
        FtStatus.DEVICE_NOT_FOUND,
    ]:
        raise FtException(result)


def get_device_info(ft_handle: FtHandle) -> Optional[ShortDeviceInfo]:
    """Get device information for an opened device.

    Args:
        ft_handle:          Handle to an opened D2XX device

    Raises:
        FtException:        In case of unexpected error

    Returns:
        ShortDeviceInfo:    Short device info
    """
    dev_type = c_uint()
    dev_id = c_uint()
    serial_number = create_string_buffer(_SERIAL_NUMBER_MAX_LEN)
    description = create_string_buffer(_DESCRIPTION_MAX_LEN)

    result: FtStatus = _get_device_info(
        ft_handle, byref(dev_type), byref(dev_id), serial_number, description, None
    )

    if result == FtStatus.OK:
        return ShortDeviceInfo(
            dev_type=DeviceType(dev_type.value),
            dev_idx=dev_id.value,
            serial_number=serial_number.value.decode("utf-8"),
            description=description.value.decode("utf-8"),
        )
    elif result in {
        FtStatus.DEVICE_NOT_FOUND,
        FtStatus.DEVICE_NOT_OPENED,
        FtStatus.INVALID_HANDLE,
    }:
        return None
    else:
        raise FtException(result)


if OS_TYPE == "Windows":

    def get_driver_version(ft_handle: FtHandle) -> DriverVersion:
        """This function returns the D2XX driver version number.

        Args:
            ft_handle:      Handle to an opened D2XX device

        Raises:
            FtException:    In case of unexpected error

        Returns:
            DriverVersion:  Tuple containing detailed driver version information
        """
        driver_version = c_uint32()

        result: FtStatus = _get_driver_version(ft_handle, byref(driver_version))

        if result != FtStatus.OK:
            raise FtException(result)

        return DriverVersion(
            build_version=driver_version.value & 0xFF,
            minor=(driver_version.value >> 8) & 0xFF,
            major=(driver_version.value >> 16) & 0xFF,
        )


def purge_buffers(ft_handle: FtHandle, purge_type_mask: BufferType) -> None:
    """This function purges receive and transmit buffers in the device.

    Args:
        ft_handle:          Handle to an opened D2XX device
        purge_type_mask:    Which buffers to purge (Tx and/or Rx)

    Raises:
        FtException:        In case of unexpected error
    """
    result: FtStatus = _purge(ft_handle, purge_type_mask)

    if result not in {
        FtStatus.OK,
        FtStatus.INVALID_HANDLE,
        FtStatus.DEVICE_NOT_FOUND,
        FtStatus.DEVICE_NOT_OPENED,
    }:
        raise FtException(result)


def reset_device(ft_handle: FtHandle) -> None:
    """This function sends a reset command to the device.

    Args:
        ft_handle:      Handle to an opened D2XX device

    Raises:
        FtException:    In case of unexpected error
    """
    result: FtStatus = _reset_device(ft_handle)

    if result not in {
        FtStatus.OK,
        FtStatus.INVALID_HANDLE,
        FtStatus.DEVICE_NOT_FOUND,
        FtStatus.DEVICE_NOT_OPENED,
    }:
        raise FtException(result)



from ctypes import POINTER, byref, c_bool, c_uint, c_uint16, c_void_p
from enum import IntEnum, auto
from typing import Final, List, NewType, Tuple

from koda import Err, Ok, Result

from . import Ft4222Exception, Ft4222Status, FtHandle, GpioTrigger
from .dll_loader import ftlib

GpioHandle = NewType("GpioHandle", FtHandle)

_GPIO_COUNT: Final[int] = 4


class Direction(IntEnum):
    OUTPUT = 0
    INPUT = auto()


class PortId(IntEnum):
    PORT_0 = 0
    PORT_1 = auto()
    PORT_2 = auto()
    PORT_3 = auto()


DirTuple = Tuple[Direction, Direction, Direction, Direction]

_init = ftlib.FT4222_GPIO_Init
_init.argtypes = [c_void_p, c_uint * 4]
_init.restype = Ft4222Status

_read = ftlib.FT4222_GPIO_Read
_read.argtypes = [c_void_p, c_uint, POINTER(c_bool)]
_read.restype = Ft4222Status

_write = ftlib.FT4222_GPIO_Write
_write.argtypes = [c_void_p, c_uint, c_bool]
_write.restype = Ft4222Status

_set_input_trigger = ftlib.FT4222_GPIO_SetInputTrigger
_set_input_trigger.argtypes = [c_void_p, c_uint, c_uint]
_set_input_trigger.restype = Ft4222Status

_get_trigger_status = ftlib.FT4222_GPIO_GetTriggerStatus
_get_trigger_status.argtypes = [c_void_p, c_uint, POINTER(c_uint16)]
_get_trigger_status.restype = Ft4222Status

_read_trigger_queue = ftlib.FT4222_GPIO_ReadTriggerQueue
_read_trigger_queue.argtypes = [
    c_void_p,
    c_uint,
    POINTER(c_uint),
    c_uint16,
    POINTER(c_uint16),
]
_read_trigger_queue.restype = Ft4222Status

_set_waveform_mode = ftlib.FT4222_GPIO_SetWaveFormMode
_set_waveform_mode.argtypes = [c_void_p, c_bool]
_set_waveform_mode.restype = Ft4222Status


def init(ft_handle: FtHandle, dirs: DirTuple) -> Result[GpioHandle, Ft4222Status]:
    """Initialize the GPIO interface of the FT4222H.

    NOTE: The GPIO interface is available on the 2nd USB interface in mode 0 or on the 4th USB interface in mode 1.

    Args:
        ft_handle:      Handle to an opened FT4222 device
        dirs:           Tuple of directions to be set for each GPIO index - [0, 1, 2, 3]

    Returns:
        Result:         Handle to initialized FT4222 device in GPIO mode
    """
    dir_array = (c_uint * _GPIO_COUNT)(*dirs)

    result: Ft4222Status = _init(ft_handle, dir_array)

    if result == Ft4222Status.OK:
        return Ok(GpioHandle(ft_handle))
    else:
        return Err(result)


def read(ft_handle: GpioHandle, port_id: PortId) -> bool:
    """Read the status of a specified GPIO pin or interrupt register.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in GPIO mode
        port_id:            GPIO port index

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        bool:               Is the port active/high?
    """
    gpio_state = c_bool()

    result: Ft4222Status = _read(ft_handle, port_id, byref(gpio_state))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return gpio_state.value


def write(ft_handle: GpioHandle, port_id: PortId, state: bool) -> None:
    """Write value to the specified GPIO pin.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in GPIO mode
        port_id:        GPIO port index
        state:          State to set

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _write(ft_handle, port_id, state)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def set_input_trigger(
    ft_handle: GpioHandle, port_id: PortId, trigger: GpioTrigger
) -> None:
    """Set software trigger conditions on the specified GPIO pin.

    This function allows developers to monitor value changes of the GPIO pins.
    Values that satisfy the trigger condition will be stored in a queue.

    For example, if 'FT4222.GpioTrigger.RISING' is set on 'PortId.GPIO_0',
    and 'PortId.GPIO_0' then changes value from 0 to 1,
    the event GPIO_TRIGGER_RISING will be recorded into the queue.

    Developers can query the queue status using 'GPIO.get_trigger_status()' and 'GPIO.read_trigger_queue()' functions.

    NOTE: This function can only set GPIO trigger conditions.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in GPIO mode
        port_id:        GPIO port index
        trigger:        Trigger type mask (using OR operator)

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_input_trigger(ft_handle, port_id, trigger)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def get_trigger_status(ft_handle: GpioHandle, port_id: PortId) -> int:
    """Get the size of trigger event queue.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in GPIO mode
        port_id:        GPIO port index

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:            Number of triggers in event queue for selected GPIO port
    """
    queue_size = c_uint16()

    result: Ft4222Status = _get_trigger_status(ft_handle, port_id, byref(queue_size))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return queue_size.value


def read_trigger_queue(
    ft_handle: GpioHandle, port_id: PortId, max_read_size: int = (2 ** 16) - 1
) -> List[GpioTrigger]:
    """Get events recorded in the trigger event queue.

    After calling this function, all events will be removed from the event queue.

    Args:
        ft_handle:                  Handle to an initialized FT4222 device in GPIO mode
        port_id:                    GPIO port index
        max_read_size:              Non-negative number of event to read from queue

    Raises:
        Ft4222Exception:            In case of unexpected device error

    Returns:
        List[FT4222.GpioTrigger]:   List of trigger events (if any)
    """
    assert (
        0 <= max_read_size < (2 ** 16)
    ), "Max. read size must be a non-negative number smaller than 2^16."

    event_buffer = (c_uint * max_read_size)()
    events_read = c_uint16()

    result: Ft4222Status = _read_trigger_queue(
        ft_handle, port_id, event_buffer, len(event_buffer), byref(events_read)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return list(map(GpioTrigger, event_buffer[: events_read.value]))


def set_waveform_mode(ft_handle: GpioHandle, enable: bool) -> None:
    """Enable or disable WaveForm Mode.

    When WaveForm mode is enabled, the device will record all GPIO status periodically.
    The peeking time depends on the system clock.
    The default setting of WaveForm mode is disabled.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in GPIO mode
        enable:         Enable WaveForm mode?

    Raises:
        Ft4222Exception:    In case of unexpected device error
    """
    result: Ft4222Status = _set_waveform_mode(ft_handle, enable)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)



from ctypes import POINTER, byref, c_char_p, c_uint, c_uint8, c_uint16, c_void_p
from enum import IntEnum, IntFlag, auto
from typing import Literal, NewType, Union, overload

from koda import Err, Ok, Result

from .. import Ft4222Exception, Ft4222Status, FtHandle
from ..dll_loader import ftlib
from . import ClkPhase, ClkPolarity

SpiSlaveRawHandle = NewType("SpiSlaveRawHandle", FtHandle)
SpiSlaveProtoHandle = NewType("SpiSlaveProtoHandle", FtHandle)
SpiSlaveHandle = Union[SpiSlaveRawHandle, SpiSlaveProtoHandle]


class IoProtocol(IntEnum):
    WITH_PROTOCOL = 0
    NO_PROTOCOL = auto()
    NO_ACK = auto()


class EventType(IntFlag):
    EVENT_RXCHAR = 8


_init = ftlib.FT4222_SPISlave_Init
_init.argtypes = [c_void_p]
_init.restype = Ft4222Status

_init_ex = ftlib.FT4222_SPISlave_InitEx
_init_ex.argtypes = [c_void_p, c_uint]
_init_ex.restype = Ft4222Status

_set_mode = ftlib.FT4222_SPISlave_SetMode
_set_mode.argtypes = [c_void_p, c_uint, c_uint]
_set_mode.restype = Ft4222Status

_get_rx_status = ftlib.FT4222_SPISlave_GetRxStatus
_get_rx_status.argtypes = [c_void_p, POINTER(c_uint16)]
_get_rx_status.restype = Ft4222Status

_read = ftlib.FT4222_SPISlave_Read
_read.argtypes = [c_void_p, POINTER(c_uint8), c_uint16, POINTER(c_uint16)]
_read.restype = Ft4222Status

_write = ftlib.FT4222_SPISlave_Write
_write.argtypes = [c_void_p, c_char_p, c_uint16, POINTER(c_uint16)]
_write.restype = Ft4222Status

_set_event_notification = ftlib.FT4222_SetEventNotification
_set_event_notification.argtypes = [c_void_p, c_uint, c_void_p]
_set_event_notification.restype = Ft4222Status


def init(ft_handle: FtHandle) -> Result[SpiSlaveProtoHandle, Ft4222Status]:
    """Initialize the FT4222H as an SPI slave.

    Default SPI Slave protocol is 'IoProtocol.WITH_PROTOCOL'.

    Args:
        ft_handle:  Handle to an open FT4222 device

    Returns:
        Result:     Handle to FT4222 device in SPI Slave mode with protocol
    """
    result: Ft4222Status = _init(ft_handle)

    if result == Ft4222Status.OK:
        return Ok(SpiSlaveProtoHandle(ft_handle))
    else:
        return Err(result)


@overload
def init_ex(
    ft_handle: FtHandle, protocol: Literal[IoProtocol.NO_PROTOCOL]
) -> Result[SpiSlaveRawHandle, Ft4222Status]:
    ...


@overload
def init_ex(
    ft_handle: FtHandle, protocol: Literal[IoProtocol.WITH_PROTOCOL, IoProtocol.NO_ACK]
) -> Result[SpiSlaveProtoHandle, Ft4222Status]:
    ...


@overload
def init_ex(
    ft_handle: FtHandle, protocol: IoProtocol
) -> Union[
    Result[SpiSlaveProtoHandle, Ft4222Status], Result[SpiSlaveRawHandle, Ft4222Status]
]:
    ...


def init_ex(
    ft_handle: FtHandle, protocol: IoProtocol
) -> Union[
    Result[SpiSlaveProtoHandle, Ft4222Status], Result[SpiSlaveRawHandle, Ft4222Status]
]:
    """Initialize the FT4222H as an SPI slave.

    Similar to 'spi.slave.init()' function, but with parameters to define the transmission protocol.

    Args:
        ft_handle:      Handle to an open FT4222 device
        protocol:       Protocol to be used for communication (if any)

    Returns:
        Result: Handle to FT4222 device in selected protocol mode
    """
    result: Ft4222Status = _init_ex(ft_handle, protocol)

    if result == Ft4222Status.OK:
        if protocol == IoProtocol.NO_PROTOCOL:
            return Ok(SpiSlaveRawHandle(ft_handle))
        else:
            return Ok(SpiSlaveProtoHandle(ft_handle))
    else:
        return Err(result)


def set_mode(
    ft_handle: SpiSlaveHandle, clk_polarity: ClkPolarity, clk_phase: ClkPhase
) -> None:
    """Set clock polarity and phase.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Slave mode
        clk_polarity:       Clock polarity (idle low/high)
        clk_phase:          Clock phase

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_mode(ft_handle, clk_polarity, clk_phase)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def get_rx_status(ft_handle: SpiSlaveHandle) -> int:
    """Get number of bytes in the receive queue.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Slave mode

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:                Number of bytes in the receive queue
    """
    rx_size = c_uint16()

    result: Ft4222Status = _get_rx_status(ft_handle, byref(rx_size))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return rx_size.value


def read(ft_handle: SpiSlaveHandle, read_byte_count: int) -> bytes:
    """Read data from the receive queue of the SPI slave device.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Slave mode
        read_byte_count:    Positive number of bytes to read from Rx queue

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        bytes:              Read data (if any)
    """
    assert (
        0 < read_byte_count < (2 ** 16)
    ), "Number of bytes to read must be positive and less than 2^16"

    read_buffer = (c_uint8 * read_byte_count)()
    bytes_read = c_uint16()

    result: Ft4222Status = _read(
        ft_handle, read_buffer, len(read_buffer), byref(bytes_read)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes(read_buffer[: bytes_read.value])


def write(ft_handle: SpiSlaveHandle, write_data: bytes) -> int:
    """Write data to the transmit queue of the SPI slave device.

    NOTE: For some reasons, support lib will append a dummy byte (0x00) at the first byte automatically.
    This additional byte exists at all of the three transfer methods.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in SPI Slave mode
        write_data:         Non-empty list of bytes to be written into Tx queue

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:                Number of bytes written into Tx queue
    """
    assert (
        0 < len(write_data) < (2 ** 16)
    ), "Data to be written must be non-empty and contain less than 2^16 bytes"

    bytes_written = c_uint16()

    result: Ft4222Status = _write(
        ft_handle, write_data, len(write_data), byref(bytes_written)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes_written.value


# FIXME: Proper typing
def set_event_notification(
    ft_handle: SpiSlaveProtoHandle, mask: EventType, param: c_void_p
) -> None:
    """Sets conditions for event notification.

    An application can use this function to setup conditions which allow a thread to block until one of the conditions is met.
    Typically, an application will create an event, call this function, and then block on the event.
    When the conditions are met, the event is set, and the application thread unblocked.
    Usually, the event is set to notify the application to check the condition.
    The application needs to check the condition again before it goes to handle the condition.
    The API is only valid when the device acts as SPI slave and SPI slave protocol is not 'IoProtocol.No_PROTOCOL'.

    Args:
        ft_handle:  Handle to an initialized FT4222 device in SPI Slave mode
        mask:       Event mask (i.e. select which events to react to)
        param:      TODO

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_event_notification(ft_handle, mask.value, param)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)



from ctypes import POINTER, byref, c_bool, c_char_p, c_uint8, c_uint16, c_void_p
from typing import NewType

from koda import Err, Ok, Result

from .. import Ft4222Exception, Ft4222Status, FtHandle
from ..dll_loader import ftlib

I2cSlaveHandle = NewType("I2cSlaveHandle", FtHandle)

# Function prototypes

_init = ftlib.FT4222_I2CSlave_Init
_init.argtypes = [c_void_p]
_init.restype = Ft4222Status

_reset = ftlib.FT4222_I2CSlave_Reset
_reset.argtypes = [c_void_p]
_reset.restype = Ft4222Status

_get_address = ftlib.FT4222_I2CSlave_GetAddress
_get_address.argtypes = [c_void_p, POINTER(c_uint8)]
_get_address.restype = Ft4222Status

_set_address = ftlib.FT4222_I2CSlave_SetAddress
_set_address.argtypes = [c_void_p, c_uint8]
_set_address.restype = Ft4222Status

_get_rx_status = ftlib.FT4222_I2CSlave_GetRxStatus
_get_rx_status.argtypes = [c_void_p, POINTER(c_uint16)]
_get_rx_status.restype = Ft4222Status

_read = ftlib.FT4222_I2CSlave_Read
_read.argtypes = [c_void_p, POINTER(c_uint8), c_uint16, POINTER(c_uint16)]
_read.restype = Ft4222Status

_write = ftlib.FT4222_I2CSlave_Write
_write.argtypes = [c_void_p, c_char_p, c_uint16, POINTER(c_uint16)]
_write.restype = Ft4222Status

_set_clock_stretch = ftlib.FT4222_I2CSlave_SetClockStretch
_set_clock_stretch.argtypes = [c_void_p, c_bool]
_set_clock_stretch.restype = Ft4222Status

_set_resp_word = ftlib.FT4222_I2CSlave_SetRespWord
_set_resp_word.argtypes = [c_void_p, c_uint8]
_set_resp_word.restype = Ft4222Status


def init(ft_handle: FtHandle) -> Result[I2cSlaveHandle, Ft4222Status]:
    """Initialized FT4222H as an I2C slave.

    Note:
        The I2C slave address is set to 0x40 after initialization.

    Args:
        ft_handle:      Handle to an opened FT4222 device

    Returns:
        Result:         Handle to initialized FT4222 device in I2C Slave mode
    """
    result: Ft4222Status = _init(ft_handle)

    if result == Ft4222Status.OK:
        return Ok(I2cSlaveHandle(ft_handle))
    else:
        return Err(result)


def reset(ft_handle: I2cSlaveHandle) -> None:
    """Reset the I2C slave device.

    This function will maintain the original I2C slave settings
    and clear all caches in the device.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Slave mode

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _reset(ft_handle)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def get_address(ft_handle: I2cSlaveHandle) -> int:
    """Get the address of the I2C slave device.

    Default address is 0x40.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Slave mode

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:            Current I2C slave address
    """
    addr = c_uint8()

    result: Ft4222Status = _get_address(ft_handle, byref(addr))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return addr.value


def set_address(ft_handle: I2cSlaveHandle, addr: int) -> None:
    """Set the address of the I2C slave device.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Slave mode
        addr:           Address to be set

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    assert (
        0 <= addr < (2 ** 7)
    ), "Device address must be an unsigned 16b integer (range 0 - 65 535)"

    result: Ft4222Status = _set_address(ft_handle, addr)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def get_rx_status(ft_handle: I2cSlaveHandle) -> int:
    """Get number of bytes in the receive queue.

    Args:
        ft_handle:  Handle to an initialized FT4222 device in I2C Slave mode

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:                Number of bytes in Rx queue
    """
    rx_size = c_uint16()

    result: Ft4222Status = _get_rx_status(ft_handle, byref(rx_size))

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return rx_size.value


def read(ft_handle: I2cSlaveHandle, read_byte_count: int) -> bytes:
    """Read data from the buffer of the I2C slave device.

    Args:
        ft_handle:          Handle to an initialized FT4222 device in I2C Slave mode
        read_byte_count:    Positive number of bytes to read

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        bytes:              Read data
    """
    assert (
        0 < read_byte_count < (2 ** 16)
    ), "Number of bytes to read must be positive and less than 2^16"

    read_buffer = (c_uint8 * read_byte_count)()
    bytes_read = c_uint16()

    result: Ft4222Status = _read(
        ft_handle, read_buffer, len(read_buffer), byref(bytes_read)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes(read_buffer[: bytes_read.value])


def write(ft_handle: I2cSlaveHandle, write_data: bytes) -> int:
    """Write data to the buffer of I2C slave device.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Slave mode
        write_data:     Non-empty list of bytes to write into Tx queue

    Raises:
        Ft4222Exception:    In case of unexpected error

    Returns:
        int:            Number of bytes written
    """
    assert (
        0 < len(write_data) < (2 ** 16)
    ), "Data to be written must be non-empty and contain less than 2^16 bytes"

    bytes_written = c_uint16()

    result: Ft4222Status = _write(
        ft_handle, write_data, len(write_data), byref(bytes_written)
    )

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)

    return bytes_written.value


def set_clock_stretch(ft_handle: I2cSlaveHandle, enable: bool) -> None:
    """Enable or disable Clock Stretch.

    The default setting of clock stretching is disabled.

    Clock stretch is as a flow-control mechanism for slaves.
    An addressed slave device may hold the clock line (SCL) low after receiving (or sending) a byte,
    indicating that it is not yet ready to process more data.
    The master that is communicating with the slave may not finish the transmission of the current bit,
    but must wait until the clock line actually goes high.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Slave mode
        enable:         Enable clock stretching?

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    result: Ft4222Status = _set_clock_stretch(ft_handle, enable)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)


def set_resp_word(ft_handle: I2cSlaveHandle, response_word: int) -> None:
    """Set the response word in case of empty Tx queue.

    Default value is 0xFF.

    This function only takes effect when Clock Stretch is disabled.
    When data is requested by an I2C master and the device is not ready to respond,
    the device will respond with a default value.

    Args:
        ft_handle:      Handle to an initialized FT4222 device in I2C Slave mode
        response_word:  Unsigned 8-bit response word to be set

    Raises:
        Ft4222Exception:    In case of unexpected error
    """
    assert (
        0 <= response_word < (2 ** 8)
    ), "The response word must be an 8b unsigned integer (range 0 - 255)"

    result: Ft4222Status = _set_resp_word(ft_handle, response_word)

    if result != Ft4222Status.OK:
        raise Ft4222Exception(result)



from typing import Generic

from pyft4222.handle import GenericProtocolHandle, StreamHandleType
from pyft4222.wrapper import Ft4222Exception, Ft4222Status
from pyft4222.wrapper.i2c.slave import (
    I2cSlaveHandle,
    get_address,
    get_rx_status,
    read,
    reset,
    set_address,
    set_clock_stretch,
    set_resp_word,
    write,
)


class I2CSlave(
    Generic[StreamHandleType],
    GenericProtocolHandle[I2cSlaveHandle, "I2CSlave", StreamHandleType],
):
    """A class encapsulating I2C Slave functions."""

    def __init__(self, ft_handle: I2cSlaveHandle, stream_handle: StreamHandleType):
        """Initialize the class with given FT4222 handle and a mode class type.

        Args:
            ft_handle:      FT4222 handle initialized in I2C Master mode
            stream_handle:  Calling stream mode handle. Used in 'uninitialize()' method.
        """
        super().__init__(ft_handle, stream_handle)

    def get_address(self) -> int:
        """Get the address of I2C Slave device.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            int:                Current I2C Slave address
        """
        if self._handle is not None:
            return get_address(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )

    def set_address(self, addr: int) -> None:
        """Set the address of I2C Slave device.

        Args:
            addr:       Device address;     range <0, 65_535>

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            if 0 <= addr < (2 ** 16):
                set_address(self._handle, addr)
            else:
                raise ValueError("addr must be in range <0, 65_535>.")
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )

    def get_rx_status(self) -> int:
        """Get number of bytes in Rx queue.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            int:                Number of bytes in Rx queue
        """
        if self._handle is not None:
            return get_rx_status(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )

    def read(self, read_byte_count: int) -> bytes:
        """Read data from the Rx queue.

        Args:
            read_byte_count:    Number of bytes to read;    range <1, 65_535>

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            bytes:              Read data
        """
        if self._handle is not None:
            if 0 < read_byte_count < (2 ** 16):
                return read(self._handle, read_byte_count)
            else:
                raise ValueError("read_byte_count must be in range <1, 65_535>.")
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )

    def write(self, write_data: bytes) -> int:
        """Write data into Tx queue.

        Args:
            write_data:     Non-empty list of bytes to write;   length <1, 65_535>

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            int:            Number of bytes written
        """
        if self._handle is not None:
            if 0 < len(write_data) < (2 ** 16):
                return write(self._handle, write_data)
            else:
                raise ValueError("write_data length must be in range <1, 65_535>.")
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )

    def set_clock_stretch(self, enable: bool) -> None:
        """Enable or disable clock stretching.

        Note:
            Clock stretching is disabled by default.

        Args:
            enable:     Enable clock stretching?

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_clock_stretch(self._handle, enable)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )

    def set_resp_word(self, response_word: int) -> None:
        """Set response word in case of empty Tx queue.

        Note:
            The default value is 0xFF.

        This function only takes effect when Clock Stretch is disabled.
        When data is requested by an I2C master and the device
        is not ready to respond, the device will respond with a default value.

        Args:
            response_word:      A new response word;    range <0, 255>

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            if 0 <= response_word < (2 ** 8):
                set_resp_word(self._handle, response_word)
            else:
                raise ValueError("response_word must be in range <0, 255>.")
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )

    def reset(self) -> None:
        """Reset the I2C Slave.

        Note:
            It is no necessary to re-initialize the controller.
            It retains all its settings, and is ready to be used again.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            reset(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "I2C Slave has been uninitialized!"
            )



from abc import ABC
from types import TracebackType
from typing import Any, ContextManager, Generic, Optional, Type, TypeVar

from pyft4222.wrapper import (
    OS_TYPE,
    Ft4222Exception,
    Ft4222Status,
    FtHandle,
    GpioTrigger,
)
from pyft4222.wrapper.common import (
    ClockRate,
    InitializedHandle,
    SwChipVersion,
    chip_reset,
    get_clock,
    get_version,
    set_clock,
    set_interrupt_trigger,
    set_suspend_out,
    set_wakeup_interrupt,
    uninitialize,
)
from pyft4222.wrapper.ftd2xx import (
    BufferType,
    DriverVersion,
    ShortDeviceInfo,
    close_handle,
    get_device_info,
    purge_buffers,
    reset_device,
)

if OS_TYPE == "Windows":
    from pyft4222.wrapper.ftd2xx import get_driver_version

HandleType = TypeVar("HandleType", bound=FtHandle)
ContextType = TypeVar("ContextType")


class GenericHandle(Generic[HandleType, ContextType], ContextManager[ContextType], ABC):
    """An abstract class encapsulating common FT4222 functions."""

    _handle: Optional[HandleType]

    def __init__(self, ft_handle: HandleType):
        """Initialize GenericHandle with given FtHandle.

        Args:
            ft_handle:      Handle to an opened FT4222 device
        """
        self._handle = ft_handle

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        if self._handle is not None:
            self.close()

        return False

    def close(self) -> None:
        """Close the current FT4222 handle.

        Note:
            A FT4222 stream must be opened again after calling this method.

        Raises:
            FtException:    In case of unexpected error
        """
        if self._handle is not None:
            close_handle(self._handle)
            self._handle = None

    def set_clock(self, clk_rate: ClockRate) -> None:
        """Set the FT4222 system clock frequency.

        Args:
            clk_rate:   Clock frequency to set

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_clock(self._handle, clk_rate)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def get_clock(self) -> ClockRate:
        """Get the current FT4222 system clock frequency.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            ClockRate:          System clock frequency
        """
        if self._handle is not None:
            return get_clock(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def set_suspend_out(self, enable: bool) -> None:
        """Enable or disable 'suspend out' function.

        By enabling the 'suspend out' function, the FT4222 will signalize
        entrance into suspend mode on GPIO2 pin.

        Warning:
            GPIO access is not possible while 'suspend out' function is enabled.

        Args:
            enable:     Enable 'suspend out' function?

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_suspend_out(self._handle, enable)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def set_wakeup_interrupt(self, enable: bool) -> None:
        """Enable or disable a wake-up/interrupt function.

        When wake-up/interrupt function is enabled, the GPIO3 acts
        as an input pin.
        While system is in normal mode, the GPIO3 acts as an interrupt pin.
        While system is in suspend mode, the GPIO3 acts as a wake-up pin.

        Warning:
            GPIO access is not possible while wake-up/interrupt is enabled.

        Args:
            enable:     Enable wake-up/interrupt function?

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_wakeup_interrupt(self._handle, enable)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def set_interrupt_trigger(self, trigger: GpioTrigger) -> None:
        """Set the trigger condition for wake-up/interrupt function.

        Note:
            By default, the trigger condition is 'GpioTrigger.RISING'.

        Warning:
            In case the device is suspended and the GPIO3 functions as interrupt,
            only 'GpioTrigger.RISING' and 'GpioTrigger.FALLING'
            modes are supported.
            The 'GpioTrigger.LEVEL_LOW' and 'GpioTrigger.LEVEL_HIGH' modes
            will react to edges instead!

        Args:
            trigger:        A wake-up/interrupt trigger to set

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            set_interrupt_trigger(self._handle, trigger)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def get_version(self) -> SwChipVersion:
        """Get version of the FT4222 chip and library.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SwChipVersion:      Version of FT4222 chip and library
        """
        if self._handle is not None:
            return get_version(self._handle)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def get_device_info(self) -> ShortDeviceInfo:
        """Get information about this opened device.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            ShortDeviceInfo:    Info about device belonging to the owned handle
        """
        if self._handle is not None:
            result = get_device_info(self._handle)
            if result is not None:
                return result
            else:
                raise Ft4222Exception(Ft4222Status.OTHER_ERROR, "Unknown error!")
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    if OS_TYPE == "Windows":

        def get_driver_version(self) -> DriverVersion:
            """Get D2XX driver version.

            Raises:
                Ft4222Exception:    In case of unexpected error

            Returns:
                DriverVersion:      Return driver version info
            """
            if self._handle is not None:
                return get_driver_version(self._handle)
            else:
                raise Ft4222Exception(
                    Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
                )

    def reset_device(self) -> None:
        """Send a reset command to the device.

        Note:
            This command closes the owned handle.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            reset_device(self._handle)
            # Handle must be closed after the device is reset
            self.close()
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def purge_buffers(self, buffer_mask: BufferType) -> None:
        """Purge the selected device buffers.

        Args:
            buffer_mask:    OR-mask selecting the buffer types to purge

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            purge_buffers(self._handle, buffer_mask)
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )

    def chip_reset(self) -> None:
        """Software reset the FT4222 device.

        Note:
            This command closes the owned handle.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        if self._handle is not None:
            chip_reset(self._handle)
            # Handle must be closed after the device is reset
            self.close()
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED, "This handle is closed!"
            )


InitializedHandleType = TypeVar("InitializedHandleType", bound=InitializedHandle)
StreamHandleType = TypeVar("StreamHandleType", bound=GenericHandle[FtHandle, Any])


class GenericProtocolHandle(
    Generic[InitializedHandleType, ContextType, StreamHandleType],
    GenericHandle[InitializedHandleType, ContextType],
    ABC,
):
    _stream_handle: StreamHandleType
    """Reference to the used stream handle"""

    def __init__(
        self, ft_handle: InitializedHandleType, stream_handle: StreamHandleType
    ):
        super().__init__(ft_handle)
        self._stream_handle = stream_handle

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        if self._handle is not None:
            self.uninitialize()

        return False

    def close(self) -> None:
        """Un-initialize and close the owned handle.

        Note:
            A new handle must be opened and initialized
            after calling this method.

        Raises:
            Ft4222Exception:    In case of unexpected error
        """
        self.uninitialize().close()

    def uninitialize(self) -> StreamHandleType:
        """Un-initialize the owned handle from the current stream mode.

        The handle can be initialized into any other supported mode.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            C:                  A class encapsulating the opened stream type

        """
        if self._handle is not None:
            new_handle = uninitialize(self._handle)
            self._stream_handle._handle = new_handle
            self._handle = None
            return self._stream_handle
        else:
            raise Ft4222Exception(
                Ft4222Status.DEVICE_NOT_OPENED,
                "Handle is already uninitialized or invalid!",
            )



"""Module encapsulating all possible FT4222 USB stream modes.

FT4222 chip can be set into one of 4 possible modes externally.
Each mode has a specific set of available USB stream interfaces.

Each class represents a specific mode of stream and encapsulates it.

Important:
    Do not instantiate classes in this module directly. Use functions
    from the ft4222 module instead.
"""

from enum import Enum, auto
from typing import Literal, Union

from koda import Ok

from pyft4222.gpio import Gpio
from pyft4222.handle import GenericHandle
from pyft4222.i2c.master import I2CMaster
from pyft4222.i2c.slave import I2CSlave
from pyft4222.spi.master import SpiMasterMulti, SpiMasterSingle
from pyft4222.spi.slave import SpiSlaveProto, SpiSlaveRaw
from pyft4222.wrapper import Ft4222Exception, Ft4222Status, FtHandle, gpio
from pyft4222.wrapper.i2c import master as i2c_master
from pyft4222.wrapper.i2c import slave as i2c_slave
from pyft4222.wrapper.spi import master as spi_master
from pyft4222.wrapper.spi import slave as spi_slave


class InterfaceType(Enum):
    """FT4222 USB interface type.

    The FT4222 chip can be set (externally) into 4 distinct modes,
    each supporting a certain functions over number of USB streams:

    +---------------+-------------+------------+------------+-------------+
    |               |   MODE 0    |   MODE 1   |   MODE 2   |   MODE 3    |
    +---------------+-------------+------------+------------+-------------+
    | Stream #0 (A) | Data Stream | SPI Master | SPI Master | Data Stream |
    | Stream #1 (B) | GPIO        | SPI Master | SPI Master | N/A         |
    | Stream #2 (C) | N/A         | SPI Master | SPI Master | N/A         |
    | Stream #3 (D) | N/A         | GPIO       | SPI Master | N/A         |
    +---------------+-------------+------------+------------+-------------+

    To select the configuration mode/interface type, you must tie the
    DCNF1, DCNF0 pins to corresponding values.

    More info can be found in the FT4222H Datasheet:
    https://ftdichip.com/wp-content/uploads/2020/07/DS_FT4222H.pdf
    """

    DATA_STREAM = auto()
    """All serial protocols are supported in this mode, excluding GPIO"""
    GPIO = auto()
    """Only GPIO is supported"""
    SPI_MASTER = auto()
    """Only SPI Master is supported"""


class ProtocolStream(GenericHandle[FtHandle, "ProtocolStream"]):
    """A class representing an open FT4222 stream in "Data Stream" mode.

    Attributes:
        tag:    Can be used to disambiguate between different stream types
    """

    tag: Literal[InterfaceType.DATA_STREAM]

    def __init__(self, ft_handle: FtHandle):
        """Initialize ProtocolHandle with given FtHandle.

        Args:
            ft_handle:          Handle to an opened FT4222 device

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            ProtocolHandle:     FT4222 USB stream handle
        """
        super().__init__(ft_handle)
        self.tag = InterfaceType.DATA_STREAM

    def init_single_spi_master(
        self,
        clk_div: spi_master.ClkDiv,
        clk_polarity: spi_master.ClkPolarity,
        clk_phase: spi_master.ClkPhase,
        sso_map: spi_master.SsoMap,
    ) -> SpiMasterSingle["ProtocolStream"]:
        """Initialize SPI Master mode, using single data line.

        Args:
            clk_div:        Divisor of system clock to create serial clock
            clk_polarity:   Serial clock polarity
            clk_phase:      Serial clock phase
            sso_map:        Slave select map

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMasterSingle:    SPI Master mode handle
        """
        if self._handle is not None:
            result = spi_master.init(
                self._handle,
                spi_master.IoMode.SINGLE,
                clk_div,
                clk_polarity,
                clk_phase,
                sso_map,
            )
            if isinstance(result, Ok):
                self._handle = None
                return SpiMasterSingle(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_dual_spi_master(
        self,
        clk_div: spi_master.ClkDiv,
        clk_polarity: spi_master.ClkPolarity,
        clk_phase: spi_master.ClkPhase,
        sso_map: spi_master.SsoMap,
    ) -> SpiMasterMulti["ProtocolStream"]:
        """Initialize SPI Master mode, using two data lines.

        Args:
            clk_div:        Divisor of system clock to create serial clock
            clk_polarity:   Serial clock polarity
            clk_phase:      Serial clock phase
            sso_map:        Slave select map

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMasterMulti:     SPI Master mode handle
        """
        if self._handle is not None:
            result = spi_master.init(
                self._handle,
                spi_master.IoMode.DUAL,
                clk_div,
                clk_polarity,
                clk_phase,
                sso_map,
            )
            if isinstance(result, Ok):
                self._handle = None
                return SpiMasterMulti(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_quad_spi_master(
        self,
        clk_div: spi_master.ClkDiv,
        clk_polarity: spi_master.ClkPolarity,
        clk_phase: spi_master.ClkPhase,
        sso_map: spi_master.SsoMap,
    ) -> SpiMasterMulti["ProtocolStream"]:
        """Initialize SPI Master mode, using four data lines.

        Args:
            clk_div:        Divisor of system clock to create serial clock
            clk_polarity:   Serial clock polarity
            clk_phase:      Serial clock phase
            sso_map:        Slave select map

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMasterMulti:     SPI Master mode handle
        """
        if self._handle is not None:
            result = spi_master.init(
                self._handle,
                spi_master.IoMode.QUAD,
                clk_div,
                clk_polarity,
                clk_phase,
                sso_map,
            )
            if isinstance(result, Ok):
                self._handle = None
                return SpiMasterMulti(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_raw_spi_slave(self) -> SpiSlaveRaw["ProtocolStream"]:
        """Initialize SPI Slave in raw mode.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiSlaveRaw:        SPI Slave raw mode handle
        """
        if self._handle is not None:
            result = spi_slave.init_ex(
                self._handle,
                spi_slave.IoProtocol.NO_PROTOCOL,
            )
            if isinstance(result, Ok):
                self._handle = None
                return SpiSlaveRaw(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_proto_spi_slave(
        self,
        proto_type: Union[
            Literal[spi_slave.IoProtocol.NO_ACK],
            Literal[spi_slave.IoProtocol.WITH_PROTOCOL],
        ],
    ) -> SpiSlaveProto["ProtocolStream"]:
        """Initialize SPI Slave in selected protocol mode.

        For protocol details, see:
        https://www.ftdichip.com/Support/Documents/AppNotes/AN_329_User_Guide_for_LibFT4222.pdf

        Protocol details can be found in chapter 3.4.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiSlaveProto:      SPI Slave protocol mode handle
        """
        if self._handle is not None:
            result = spi_slave.init_ex(self._handle, proto_type)
            if isinstance(result, Ok):
                self._handle = None
                return SpiSlaveProto(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_i2c_master(self, kbps: int) -> I2CMaster["ProtocolStream"]:
        """Initialize I2C Master.

        Args:
            kbps:   I2C clock/transmission speed in kbit/s (60 - 3400)

        Raises:
            Ft4222Exception:    In case of an unexpected error

        Returns:
            I2CMaster:          I2C Master handle
        """
        if self._handle is not None:
            result = i2c_master.init(self._handle, kbps)
            if isinstance(result, Ok):
                self._handle = None
                return I2CMaster(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_i2c_slave(self) -> I2CSlave["ProtocolStream"]:
        """Initialize I2C Slave.

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            I2CSlave:           I2C Slave handle
        """
        if self._handle is not None:
            result = i2c_slave.init(self._handle)
            if isinstance(result, Ok):
                self._handle = None
                return I2CSlave(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)


class GpioStream(GenericHandle[FtHandle, "GpioStream"]):
    """A class representing an open FT4222 stream in "GPIO" mode.

    Attributes:
        tab:    Can be used to disambiguate between different stream types
    """

    tag: Literal[InterfaceType.GPIO]

    def __init__(self, ft_handle: FtHandle):
        """Initialize GpioHandle with given FtHandle.

        Args:
            ft_handle:          Handle to an opened FT4222 device
        """
        super().__init__(ft_handle)
        self.tag = InterfaceType.GPIO

    def init_gpio(self, dirs: gpio.DirTuple) -> Gpio["GpioStream"]:
        """Initialize GPIO.

        Args:
            dirs:       A 4-tuple selecting the direction of associated GPIO pin

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            Gpio:               GPIO handle
        """
        if self._handle is not None:
            result = gpio.init(self._handle, dirs)
            if isinstance(result, Ok):
                self._handle = None
                return Gpio(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)


class SpiStream(GenericHandle[FtHandle, "SpiStream"]):
    """A class representing an open FT4222 stream in "SPI Master" mode.

    Attributes:
        tag:    Can be used to disambiguate between different stream types

    """

    tag: Literal[InterfaceType.SPI_MASTER]

    def __init__(self, ft_handle: FtHandle):
        """Initialize SpiMasterHandle with given FtHandle.

        Args:
            ft_handle:          Handle to an opened FT4222 device

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMasterHandle
        """
        super().__init__(ft_handle)
        self.tag = InterfaceType.SPI_MASTER

    def init_single_spi_master(
        self,
        clk_div: spi_master.ClkDiv,
        clk_polarity: spi_master.ClkPolarity,
        clk_phase: spi_master.ClkPhase,
        sso_map: spi_master.SsoMap,
    ) -> SpiMasterSingle["SpiStream"]:
        """Initialize SPI Master mode, using single data line.

        Args:
            clk_div:        Divisor of system clock to create serial clock
            clk_polarity:   Serial clock polarity
            clk_phase:      Serial clock phase
            sso_map:        Slave select map

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMasterSingle:    SPI Master mode handle
        """
        if self._handle is not None:
            result = spi_master.init(
                self._handle,
                spi_master.IoMode.SINGLE,
                clk_div,
                clk_polarity,
                clk_phase,
                sso_map,
            )
            if isinstance(result, Ok):
                self._handle = None
                return SpiMasterSingle(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_dual_spi_master(
        self,
        clk_div: spi_master.ClkDiv,
        clk_polarity: spi_master.ClkPolarity,
        clk_phase: spi_master.ClkPhase,
        sso_map: spi_master.SsoMap,
    ) -> SpiMasterMulti["SpiStream"]:
        """Initialize SPI Master mode, using two data lines.

        Args:
            clk_div:        Divisor of system clock to create serial clock
            clk_polarity:   Serial clock polarity
            clk_phase:      Serial clock phase
            sso_map:        Slave select map

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMasterMulti:     SPI Master mode handle
        """
        if self._handle is not None:
            result = spi_master.init(
                self._handle,
                spi_master.IoMode.DUAL,
                clk_div,
                clk_polarity,
                clk_phase,
                sso_map,
            )
            if isinstance(result, Ok):
                self._handle = None
                return SpiMasterMulti(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)

    def init_quad_spi_master(
        self,
        clk_div: spi_master.ClkDiv,
        clk_polarity: spi_master.ClkPolarity,
        clk_phase: spi_master.ClkPhase,
        sso_map: spi_master.SsoMap,
    ) -> SpiMasterMulti["SpiStream"]:
        """Initialize SPI Master mode, using four data lines.

        Args:
            clk_div:        Divisor of system clock to create serial clock
            clk_polarity:   Serial clock polarity
            clk_phase:      Serial clock phase
            sso_map:        Slave select map

        Raises:
            Ft4222Exception:    In case of unexpected error

        Returns:
            SpiMasterMulti:     SPI Master mode handle
        """
        if self._handle is not None:
            result = spi_master.init(
                self._handle,
                spi_master.IoMode.QUAD,
                clk_div,
                clk_polarity,
                clk_phase,
                sso_map,
            )
            if isinstance(result, Ok):
                self._handle = None
                return SpiMasterMulti(result.val, self)
            else:
                raise Ft4222Exception(result.val)
        else:
            raise Ft4222Exception(Ft4222Status.INVALID_HANDLE)



import platform
from ctypes import c_void_p
from enum import IntEnum, IntFlag, auto
from typing import Final, NewType, Optional

OS_TYPE: Final = platform.system()

FtHandle = NewType("FtHandle", c_void_p)


class FtStatus(IntEnum):
    """Class representing a D2XX 'FT_RESULT' enum."""

    OK = 0
    INVALID_HANDLE = auto()
    DEVICE_NOT_FOUND = auto()
    DEVICE_NOT_OPENED = auto()
    IO_ERROR = auto()
    INSUFFICIENT_RESOURCES = auto()
    INVALID_PARAMETER = auto()
    INVALID_BAUD_RATE = auto()
    DEVICE_NOT_OPENED_FOR_ERASE = auto()
    DEVICE_NOT_OPENED_FOR_WRITE = auto()
    FAILED_TO_WRITE_DEVICE = auto()
    EEPROM_READ_FAILED = auto()
    EEPROM_WRITE_FAILED = auto()
    EEPROM_ERASE_FAILED = auto()
    EEPROM_NOT_PRESENT = auto()
    EEPROM_NOT_PROGRAMMED = auto()
    INVALID_ARGS = auto()
    NOT_SUPPORTED = auto()
    OTHER_ERROR = auto()
    DEVICE_LIST_NOT_READY = auto()


class FtException(Exception):
    """A class wrapping the D2XX driver exceptions.

    Attributes:
        status:     D2XX result enum
        msg:        Optional human-readable message
    """

    status: FtStatus
    msg: Optional[str]

    def __init__(self, status: FtStatus, msg: Optional[str] = None):
        super().__init__()
        self.status = status
        self.msg = msg

    def __str__(self) -> str:
        return f"""
Exception during call to FTD2XX library.
FT return code: {self.status.name}
Message: {self.msg}
        """


class Ft4222Status(IntEnum):
    """Enum representing the 'FT4222_STATUS' enum from ft4222 library.

    This enum 'extends' the 'FT_STATUS' enum.
    """

    OK = 0
    INVALID_HANDLE = auto()
    DEVICE_NOT_FOUND = auto()
    DEVICE_NOT_OPENED = auto()
    IO_ERROR = auto()
    INSUFFICIENT_RESOURCES = auto()
    INVALID_PARAMETER = auto()
    INVALID_BAUD_RATE = auto()
    DEVICE_NOT_OPENED_FOR_ERASE = auto()
    DEVICE_NOT_OPENED_FOR_WRITE = auto()
    FAILED_TO_WRITE_DEVICE = auto()
    EEPROM_READ_FAILED = auto()
    EEPROM_WRITE_FAILED = auto()
    EEPROM_ERASE_FAILED = auto()
    EEPROM_NOT_PRESENT = auto()
    EEPROM_NOT_PROGRAMMED = auto()
    INVALID_ARGS = auto()
    NOT_SUPPORTED = auto()
    OTHER_ERROR = auto()
    DEVICE_LIST_NOT_READY = auto()

    # FtStatus extending message
    DEVICE_NOT_SUPPORTED = 1000
    CLK_NOT_SUPPORTED = auto()
    VENDOR_CMD_NOT_SUPPORTED = auto()
    IS_NOT_SPI_MODE = auto()
    IS_NOT_I2C_MODE = auto()
    IS_NOT_SPI_SINGLE_MODE = auto()
    IS_NOT_SPI_MULTI_MODE = auto()
    WRONG_I2C_ADDR = auto()
    INVALID_FUNCTION = auto()
    INVALID_POINTER = auto()
    EXCEEDED_MAX_TRANSFER_SIZE = auto()
    FAILED_TO_READ_DEVICE = auto()
    I2C_NOT_SUPPORTED_IN_THIS_MODE = auto()
    GPIO_NOT_SUPPORTED_IN_THIS_MODE = auto()
    GPIO_EXCEEDED_MAX_PORTNUM = auto()
    GPIO_WRITE_NOT_SUPPORTED = auto()
    GPIO_PULLUP_INVALID_IN_INPUTMODE = auto()
    GPIO_PULLDOWN_INVALID_IN_INPUTMODE = auto()
    GPIO_OPENDRAIN_INVALID_IN_OUTPUTMODE = auto()
    INTERRUPT_NOT_SUPPORTED = auto()
    GPIO_INPUT_NOT_SUPPORTED = auto()
    EVENT_NOT_SUPPORTED = auto()
    FUN_NOT_SUPPORT = auto()

    @classmethod
    def from_ft_status(cls, ft_status: FtStatus) -> "Ft4222Status":
        return Ft4222Status(int(ft_status))


class Ft4222Exception(Exception):
    """A class wrapping the FT4222 driver exceptions.

    Params:
        status:     FT4222 result enum
        msg:        Optional human-readable message
    """

    status: Ft4222Status
    msg: Optional[str]

    def __init__(self, status: Ft4222Status, msg: Optional[str] = None):
        super().__init__()
        self.status = status
        self.msg = msg

    def __str__(self) -> str:
        return f"""
Exception during call to LibFT4222 library.
FT return code: {self.status.name}
Message: {self.msg}
        """


class GpioTrigger(IntFlag):
    """Enum representing possible GPIO triggers types."""

    RISING = 0x01
    """Trigger on rising edge."""
    FALLING = 0x02
    """Trigger on falling edge."""
    LEVEL_HIGH = 0x04
    """Trigger on high level."""
    LEVEL_LOW = 0x08
    """Trigger on low level."""
